import click
import itertools
import re


def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


def flatmap(func, items):
    return itertools.chain.from_iterable(map(func, items))


def ascii_encode(non_compatible_string):
    """Primarily used for ensuring terminal display compatibility"""
    if non_compatible_string:
        return non_compatible_string.encode("ascii", errors="ignore").decode("ascii")
    else:
        return ""


def pull(nested_dict):
    if "type" in nested_dict and "inputs" not in nested_dict:
        return nested_dict
    else:
        inputs = {}
        if "type" in nested_dict and "inputs" in nested_dict:
            for param, input in list(nested_dict["inputs"].items()):
                inputs[str(param)] = pull(input)
            return inputs
        else:
            return nested_dict


def regex_manifest(protocol, input):
    """Special input types, gets updated as more input types are added"""
    if "type" in input and input["type"] == "choice":
        if "options" in input:
            pattern = "\[(.*?)\]"
            match = re.search(pattern, str(input["options"]))
            if not match:
                click.echo(
                    'Error in %s: input type "choice" options must '
                    'be in the form of: \n[\n  {\n  "value": '
                    '<choice value>, \n  "label": <choice label>\n  '
                    "},\n  ...\n]" % protocol["name"]
                )
                raise RuntimeError
        else:
            click.echo(
                f"Must have options for 'choice' input type. Error in: {protocol['name']}"
            )
            raise RuntimeError


def iter_json(manifest):
    all_types = {}
    try:
        protocol = manifest["protocols"]
    except TypeError:
        raise RuntimeError(
            "Error: Your manifest.json file doesn't contain "
            "valid JSON and cannot be formatted."
        )
    for protocol in manifest["protocols"]:
        types = {}
        for param, input in list(protocol["inputs"].items()):
            types[param] = pull(input)
            if isinstance(input, dict):
                if input["type"] == "group" or input["type"] == "group+":
                    for i, j in list(input.items()):
                        if isinstance(j, dict):
                            for k, l in list(j.items()):
                                regex_manifest(protocol, l)
                else:
                    regex_manifest(protocol, input)
        all_types[protocol["name"]] = types
    return all_types


def by_well(datasets, well):
    return [
        datasets[reading].props["data"][well][0] for reading in list(datasets.keys())
    ]


def makedirs(name, mode=None, exist_ok=False):
    """Forward ports `exist_ok` flag for Py2 makedirs. Retains mode defaults"""
    from os import makedirs

    mode = mode if mode is not None else 0o777
    makedirs(name, mode, exist_ok)


def is_valid_jwt_token(token: str):
    regex = r"Bearer ([a-zA-Z0-9_=]+)\.([a-zA-Z0-9_=]+)\.([a-zA-Z0-9_\-\+\/=]*)"
    return re.fullmatch(regex, token) is not None



from autoprotocol.container_type import ContainerType
from transcriptic import dataset as get_dataset

try:
    import plotly as py
    import pandas
    import matplotlib.pyplot as plt
    import plotly.tools as tls
    import numpy as np
except ImportError:
    raise ImportError(
        "Please run `pip install transcriptic[analysis] if you "
        "would like to use the Transcriptic analysis module."
    )


class _PlateRead(object):

    """
    A PlateRead object generalizes the parsing of datasets derived from the
    plate reader for easy statistical analysis and visualization.

    Refer to the Absorbance, Fluorescence and Luminescence objects for more
    information.

    """

    def __init__(
        self,
        op_type,
        dataset,
        group_labels,
        group_wells=None,
        control_reading=None,
        name=None,
    ):
        self.name = name
        self.dataset = dataset
        self.control_reading = control_reading
        self.op_type = op_type
        if self.op_type not in ["absorbance", "fluorescence", "luminescence"]:
            raise RuntimeError("Data given is not from a spectrophotometry operation.")
        if self.op_type != (self.dataset.attributes["instruction"]["operation"]["op"]):
            raise RuntimeError(f"Data given is not a {op_type} operation.")

        # Populate measurement params
        measure_params_dict = dict()
        measure_params_dict["reader"] = self.dataset.attributes["warp"]["device_id"]
        dataset_op = self.dataset.attributes["instruction"]["operation"]
        if self.op_type == "absorbance":
            measure_params_dict["wavelength"] = (
                dataset_op["wavelength"].split(":")[0] + "nm"
            )
        if self.op_type == "fluorescence":
            measure_params_dict["wavelength"] = (
                f"excitation: {dataset_op['excitation'].split(':')[0] + 'nm'} "
                f"emission: {dataset_op['emission'].split(':')[0] + 'nm'}"
            )
        if self.op_type == "luminescence":
            measure_params_dict["wavelength"] = ""

        self.params = measure_params_dict

        # Populate plate field
        plate_info_dict = dict()
        plate_info_dict["id"] = self.dataset.attributes["container"]["id"]
        plate_info_dict["col_count"] = self.dataset.attributes["container"][
            "container_type"
        ]["col_count"]
        plate_info_dict["well_count"] = self.dataset.attributes["container"][
            "container_type"
        ]["well_count"]
        self.params["plate"] = plate_info_dict

        # Get dataset and parse into DataFrame
        data_dict = get_dataset(self.dataset.attributes["id"])
        self.df = pandas.DataFrame()
        well_count = self.dataset.attributes["container"]["container_type"][
            "well_count"
        ]
        col_count = self.dataset.attributes["container"]["container_type"]["col_count"]
        # If no group well list specified, default to including all well data
        # values in one group
        if not group_wells:
            self.df = pandas.DataFrame(
                [x[0] for x in list(data_dict.values())], columns=[group_labels[0]]
            )
        # If given list of all int, assume one group with all wells in list
        elif all(isinstance(i, int) for i in group_wells):
            if len(group_wells) > len(data_dict):
                raise ValueError("Sum of group lengths exceeds total no. of wells.")
            wells = [
                ContainerType.humanize_static(_, well_count, col_count).lower()
                for _ in group_wells
            ]
            if not all(_ in data_dict for _ in wells):
                raise ValueError(f"Not all wells {wells} are in dataset {data_dict}.")

            self.df = pandas.DataFrame(
                [data_dict[_][0] for _ in wells], columns=[group_labels[0]]
            )
        elif all(isinstance(i, list) for i in group_wells):
            if group_wells and (sum([len(i) for i in group_wells]) > len(data_dict)):
                raise ValueError("Sum of group lengths exceeds total no. of wells.")
            for (idx, well_list) in enumerate(group_wells):
                wells = [
                    ContainerType.humanize_static(_, well_count, col_count).lower()
                    for _ in well_list
                ]
                if not all(_ in data_dict for _ in wells):
                    raise ValueError(
                        f"Not all wells {wells} are in dataset {data_dict}."
                    )
                col = pandas.DataFrame(
                    [data_dict[_][0] for _ in wells], columns=[group_labels[0]]
                )
                # if group_well members are of different lengths,
                # concat automatically pads resultant DataFrame with NaN
                self.df = pandas.concat([self.df, col], axis=1)
        else:
            raise ValueError(
                "Format Error: Group Well List should be a list of list of \
                 wells in robot format"
            )

        # If control absorbance object specified, create df_abj variable by
        # subtracting control df from original
        if control_reading:
            self.df_adj = self.df - control_reading.df

        self.cv = self.df.std() / (self.df.mean() * 100)

    def plot(self, mpl=True, plot_type="box", **plt_kwargs):
        """
        Parameters
        ----------
        mpl : boolean, optional
            Set to True to render a matplotlib plot, otherwise a Plotly plot
            is rendered
        plot_type : {"box", "bar", "line", "hist"}, optional
            Type of plot to render
        \**plot_kwargs : dict, optional
            Optional dictionary of specifications for your plot type of choice
        """
        py.offline.init_notebook_mode()

        mpl_fig, ax = plt.subplots()
        nl = "\n" if mpl else "<br>"
        ax.set_ylabel(self.op_type + nl + self.params["wavelength"])
        self.df.plot(kind=plot_type, ax=ax)
        labels = [item.label.get_text() for item in ax.xaxis.get_major_ticks()]
        if mpl:
            return None
        else:
            if not plt_kwargs:
                plt_kwargs = {
                    "layout": {
                        "xaxis": {
                            "tickmode": "array",
                            "ticktext": labels,
                            "tickvals": list(range(1, len(labels) + 1)),
                            "tickangle": 0,
                            "tickfont": {"size": 10},
                        }
                    }
                }
            pyfig = tls.mpl_to_plotly(mpl_fig)
            pyfig.update(plt_kwargs)
            return py.offline.iplot(pyfig)


class Absorbance(_PlateRead):

    """
    An Absorbance object parses a dataset object and provides functions for
    easy statistical analysis and visualization.

    Parameters
    ----------

    dataset: dataset
        Single dataset selected from datasets object
    group_labels: list[str]
        Labels for each of the respective groups
    group_wells: list[int]
        List of list of wells (robot form) belonging to each group in order.
        E.g. [[1,3,5],[2,4,6]]
    control_abs: Absorbance object, optional
        Absorbance object of water/control blank. If specified, will create
        adjusted dataframe df_adj by subtracting from existing df
    name: str, optional
        Name of absorbance object. Used in plotting functions

    """

    def __init__(
        self, dataset, group_labels, group_wells=None, control_abs=None, name=None
    ):
        _PlateRead.__init__(
            self, "absorbance", dataset, group_labels, group_wells, control_abs, name
        )

    def beers_law(self, conc_list=None, use_adj=True, **kwargs):
        """ "
        Apply Beer-Lambert's law to a series of absorbance readings and get
        an estimation of the linearity between the absorbance and concentration
        values.

        Parameters
        ----------

        conc_list: list[double], optional
            List of concentrations of dye used
        use_adj: Boolean, optional
            Booelan option which determines if the adjusted absorbance readings
            are used
        \**plot_kwargs : dict
            Optional dictionary of specifications for your plot type of choice
        """
        if "title" not in kwargs:
            if self.name:
                kwargs["title"] = f"Beer's Law ({self.name})"
            else:
                kwargs["title"] = "Beer's Law"
        if "yerr" not in kwargs:
            kwargs["yerr"] = self.df.std()

        # Use df_adj for beer's law if control abs object was given
        if use_adj and self.control_reading:
            dataf = self.df_adj
        else:
            dataf = self.df
        # Use default labels if concentration not provided
        if not conc_list:
            if "xlim" not in kwargs:
                kwargs["xlim"] = (-1, len(dataf.mean()))
            dataf.mean().plot(**kwargs)
        else:
            plot_obj = pandas.DataFrame(
                {"values": dataf.mean(), "conc": np.asarray(conc_list)}
            )
            result = np.polyfit(plot_obj["conc"], plot_obj["values"], 1, full=True)
            gradient, intercept = result[0]
            mpl_fig, ax = plt.subplots()
            plot_obj.plot(x="conc", y="values", kind="scatter", ax=ax, **kwargs)
            plt.plot(plot_obj["conc"], gradient * plot_obj["conc"] + intercept, "-")
            ax.set_ylabel("Absorbance " + self.params["wavelength"])

            # Calculate R^2 from residuals
            ss_res = result[1]
            ss_tot = np.sum(np.square((plot_obj["values"] - plot_obj["values"].mean())))
            print(f"{self.name} R^2: {1 - ss_res // ss_tot}")


class Fluorescence(_PlateRead):

    """
    An Fluorescence object parses a dataset object and provides functions for
    easy statistical analysis and visualization.

    Parameters
    ----------

    dataset: dataset
        Single dataset selected from datasets object
    group_labels: list[str]
        Labels for each of the respective groups
    group_wells: list[int]
        List of list of wells (robot form) belonging to each group in order.
        E.g. [[1,3,5],[2,4,6]]
    control_fluor: Fluorescence object, optional
        Fluorescence object of water/control blank. If specified, will create
        adjusted dataframe df_adj by subtracting from existing df
    name: str, optional
        Name of fluorescence object. Used in plotting functions

    """

    def __init__(
        self, dataset, group_labels, group_wells=None, control_fluor=None, name=None
    ):
        _PlateRead.__init__(
            self,
            "fluorescence",
            dataset,
            group_labels,
            group_wells,
            control_fluor,
            name,
        )


class Luminescence(_PlateRead):

    """
    An Luminescence object parses a dataset object and provides functions for
    easy statistical analysis and visualization.

    Parameters
    ----------

    dataset: dataset
        Single dataset selected from datasets object
    group_labels: list[str]
        Labels for each of the respective groups
    group_wells: list[int]
        List of list of wells (robot form) belonging to each group in order.
        E.g. [[1,3,5],[2,4,6]]
    control_lumi: Luminescence object, optional
        Luminescence object of water/control blank. If specified, will create
        adjusted dataframe df_adj by subtracting from existing df
    name: str, optional
        Name of luminescence object. Used in plotting functions

    """

    def __init__(
        self, dataset, group_labels, group_wells=None, control_lumi=None, name=None
    ):
        _PlateRead.__init__(
            self, "luminescence", dataset, group_labels, group_wells, control_lumi, name
        )


def compare_standards(pr_obj, std_pr_obj):
    """
    Compare a sample plate read object with a standard plate read object to get
    measures such as the Root-Mean-Square-Error (RMSE) and
    Coefficient-of-Variation (CV).


    Parameters
    ----------

    pr_obj: _PlateRead
        Sample plate read object
    std_pr_obj: _PlateRead
        Standard plate read object
    """
    # Compare against mean of standard absorbance
    # Check to ensure CVs are at least 2 apart
    for indx in range(len(pr_obj.cv)):
        cv_ratio = pr_obj.cv.iloc[indx] // std_pr_obj.cv.iloc[indx]
        if cv_ratio < 2:
            print(
                f"Warning for {pr_obj.cv.index[indx]}: Sample CV is only "
                f"{cv_ratio} times that of Standard CV. RMSE may be inaccurate."
            )
    # RMSE (normalized wrt to standard mean)
    RMSE = (
        np.sqrt(np.square(pr_obj.df - std_pr_obj.df.mean()).mean())
        / std_pr_obj.df.mean()
        * 100
    )
    RMSE = pandas.DataFrame(RMSE, columns=["RMSE % (normalized to standard mean)"])

    sampleVariance = pandas.DataFrame(pr_obj.df.var(), columns=["Sample Variance"])
    sampleCV = pandas.DataFrame(pr_obj.cv, columns=["Sample (%) CV"])

    try:
        # pylint: disable=import-error
        from IPython.display import HTML, display

        if pr_obj.name:
            display(HTML(f"<b>Standards Comparison ({pr_obj.name})</b>"))
        display(sampleVariance)
        display(sampleCV)
        display(RMSE)
    except:
        # If IPython module is not present or unable to show, print results
        print(sampleVariance)
        print(sampleCV)
        print(RMSE)



"""
Big dumb file of routes, please do not add any logic into this file
Note: {api_root} and {org_id} are automatically supplied in the Connection.get_route and do not need to be specified
"""


def create_project(api_root, org_id):
    return "{api_root}/{org_id}".format(**locals())


def delete_project(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}".format(**locals())


def archive_project(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}".format(**locals())


def get_project(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}".format(**locals())


def get_projects(api_root, org_id):
    return "{api_root}/{org_id}/?q=&per_page=500".format(**locals())


def get_project_runs(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}/runs".format(**locals())


def create_package(api_root, org_id):
    return "{api_root}/{org_id}/packages".format(**locals())


def delete_package(api_root, org_id, package_id):
    return "{api_root}/{org_id}/packages/{package_id}".format(**locals())


def get_package(api_root, org_id, package_id):
    return "{api_root}/{org_id}/packages/{package_id}".format(**locals())


def get_packages(api_root, org_id):
    return "{api_root}/{org_id}/packages/".format(**locals())


def get_protocols(api_root, org_id):
    return "{api_root}/{org_id}/protocols".format(**locals())


def launch_protocol(api_root, org_id, protocol_id):
    return "{api_root}/{org_id}/protocols/{protocol_id}/launch".format(**locals())


def get_launch_request(api_root, org_id, protocol_id, launch_request_id):
    return (
        "{api_root}/{org_id}/protocols/{protocol_id}/launch/{launch_request_id}".format(
            **locals()
        )
    )


def post_release(api_root, org_id, package_id):
    return "{api_root}/{org_id}/packages/{package_id}/releases/".format(**locals())


def get_release_status(api_root, org_id, package_id, release_id, timestamp):
    return "{api_root}/{org_id}/packages/{package_id}/releases/{release_id}?_={timestamp}".format(
        **locals()
    )


def query_kits(api_root, query):
    return "{api_root}/_commercial/kits?q={query}&per_page=1000&full_json=true".format(
        **locals()
    )


def query_resources(api_root, query):
    return "{api_root}/_commercial/resources?q={query}&per_page=1000".format(**locals())


def query_inventory(api_root, org_id, query, page=0):
    return "{api_root}/{org_id}/inventory/samples?q={query}&per_page=75&page={page}".format(
        **locals()
    )


def get_quick_launch(api_root, org_id, project_id, quick_launch_id):
    return (
        "{api_root}/{org_id}/{project_id}/runs/quick_launch/{quick_launch_id}".format(
            **locals()
        )
    )


def create_quick_launch(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}/runs/quick_launch".format(**locals())


def resolve_quick_launch_inputs(api_root, org_id, project_id, quick_launch_id):
    return (
        "{api_root}/{org_id}/{project_id}/runs/quick_launch/"
        "{quick_launch_id}/resolve_inputs".format(**locals())
    )


def login(api_root):
    return "{api_root}/users/sign_in".format(**locals())


def get_organizations(api_root):
    return "{api_root}/organizations".format(**locals())


def get_organization(api_root, org_id):
    return "{api_root}/{org_id}".format(**locals())


def deref_route(api_root, obj_id):
    return "{api_root}/-/{obj_id}".format(**locals())


def analyze_run(api_root, org_id):
    return "{api_root}/{org_id}/analyze_run".format(**locals())


def analyze_launch_request(api_root, org_id):
    return "{api_root}/{org_id}/analyze_run".format(**locals())


def submit_run(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}/runs".format(**locals())


def submit_launch_request(api_root, org_id, project_id):
    return "{api_root}/{org_id}/{project_id}/runs".format(**locals())


def dataset_short(api_root, data_id):
    return "{api_root}/datasets/{data_id}.json".format(**locals())


def dataset(api_root, data_id, key):
    return "{api_root}/datasets/{data_id}.json?key={key}".format(**locals())


def datasets(api_root, org_id, project_id, run_id):
    return "{api_root}/{org_id}/{project_id}/runs/{run_id}/data".format(**locals())


def data_object(api_root, id):
    return "{api_root}/api/data_objects/{id}".format(**locals())


def data_objects(api_root, dataset_id):
    return "{api_root}/api/data_objects?filter[dataset_id]={dataset_id}".format(
        **locals()
    )


def get_uploads(api_root, key):
    return "{api_root}/upload/url_for?key={key}".format(**locals())


def upload(api_root):
    return "{api_root}/api/uploads".format(**locals())


def upload_datasets(api_root):
    return "{api_root}/api/datasets".format(**locals())


def modify_aliquot_properties(api_root, aliquot_id):
    return "{api_root}/api/aliquots/{aliquot_id}/modify_properties".format(**locals())


def preview_protocol(api_root):
    return "{api_root}/runs/preview".format(**locals())


def preview_protocol_embed(api_root, preview_id):
    return "{api_root}/runs/preview/{preview_id}.embed".format(**locals())


def view_data(api_root, data_id):
    return "{api_root}/datasets/{data_id}.embed".format(**locals())


def view_run(api_root, org_id, project_id, run_id):
    return "{api_root}/{org_id}/{project_id}/runs/{run_id}.embed".format(**locals())


def view_instruction(api_root, org_id, project_id, run_id, instruction_id):
    return "{api_root}/{org_id}/{project_id}/runs/{run_id}/instructions/{instruction_id}.embed".format(
        **locals()
    )


def view_raw_image(api_root, data_id):
    return "{api_root}/-/{data_id}.raw".format(**locals())


def get_data_zip(api_root, data_id):
    return "{api_root}/-/{data_id}.zip".format(**locals())


def monitoring_data(
    api_root, data_type, instruction_id, grouping=None, start_time=None, end_time=None
):
    base_route = (
        "{api_root}/sensor_data/{data_type}?instruction_id={instruction_id}".format(
            **locals()
        )
    )
    if grouping:
        base_route += "&grouping={grouping}".format(**locals())
    if start_time:
        base_route += "&start_time={start_time}".format(**locals())
    if end_time:
        base_route += "&end_time={end_time}".format(**locals())
    return base_route


def get_payment_methods(api_root, org_id):
    return "{api_root}/{org_id}/payment_methods".format(**locals())



import base64
from email.utils import formatdate
from Crypto.Hash import SHA256
from httpsig.utils import HttpSigException
from requests.auth import AuthBase
from httpsig.requests_auth import HTTPSignatureAuth


class StrateosSign(AuthBase):
    """Signs requests"""

    def __init__(self, email, secret):
        self.email = email
        self.secret = secret
        headers = ["(request-target)", "Date", "Host"]
        body_headers = ["Digest", "Content-Length"]
        try:
            self.auth = HTTPSignatureAuth(
                key_id=self.email,
                algorithm="rsa-sha256",
                headers=headers,
                secret=self.secret,
            )
            self.body_auth = HTTPSignatureAuth(
                key_id=self.email,
                algorithm="rsa-sha256",
                headers=headers + body_headers,
                secret=self.secret,
            )
        except HttpSigException:
            raise ValueError(
                "Could not parse the specified RSA Key, ensure it "
                "is a PRIVATE key in PEM format"
            )

    def __call__(self, request):
        if "Date" not in request.headers:
            request.headers["Date"] = formatdate(
                timeval=None, localtime=False, usegmt=True
            )

        if request.method.upper() in ("PUT", "POST", "PATCH"):
            digest = SHA256.new(request.body.encode()).digest()
            sha = base64.b64encode(digest).decode("ascii")
            request.headers["Digest"] = f"SHA-256={sha}"
            return self.body_auth(request)

        return self.auth(request)


class BearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = self.token
        return r



from copy import deepcopy
from io import StringIO
from operator import itemgetter
from requests.exceptions import ReadTimeout
import json
import requests
import warnings

try:
    import pandas as pd
except ImportError:
    raise ImportError(
        "Please run `pip install transcriptic[jupyter] if you "
        "would like to use Transcriptic objects."
    )


def _check_api(obj_type):
    from transcriptic import api

    if not api:
        raise RuntimeError(
            f"You have to be logged in to be able to create {obj_type} objects"
        )
    return api


class _BaseObject(object):
    """Base object which other objects inherit from"""

    # TODO: Inherit more stuff from here. Need to ensure web has unified fields for jupyter
    def __init__(self, obj_type, obj_id, attributes, connection=None):
        # If attributes and connection are explicitly provided, just return and not do any smart parsing
        if attributes and connection:
            self.connection = connection
            self.attributes = attributes
        else:
            if not connection:
                self.connection = _check_api(obj_type)
            else:
                self.connection = connection
            (self.id, self.name) = self.load_object(obj_type, obj_id)
            if not attributes:
                self.attributes = self.connection._get_object(self.id, obj_type)
            else:
                self.attributes = attributes

    def load_object(self, obj_type, obj_id):
        """Find and match object by name"""
        # TODO: Remove the try/except statement and properly handle cases where objects are not found
        # TODO: Fix `datasets` route since that only returns non-analysis objects
        try:
            objects = getattr(self.connection, obj_type + "s")()
        except:
            return (obj_id, str(obj_id))
        matched_objects = []
        for obj in objects:
            # Special case here since we use both 'name' and 'title' for object names
            if "name" in obj:
                if obj_id == obj["name"] or obj_id == obj["id"]:
                    matched_objects.append((obj["id"], obj["name"]))
            if "title" in obj:
                if obj_id == obj["title"] or obj_id == obj["id"]:
                    matched_objects.append((obj["id"], obj["title"]))
        if len(matched_objects) == 0:
            raise TypeError(f"{obj_id} is not found in your {obj_type}s.")
        elif len(matched_objects) == 1:
            return matched_objects[0]
        else:
            print(
                f"More than 1 match found. Defaulting to the first match: {matched_objects[0]}"
            )
            return matched_objects[0]


class Project(_BaseObject):
    """
    A Project object contains helper methods for managing your runs. You can view the runs associated with this project
    as well as submit runs to the project.

    Example Usage:

    .. code-block:: python

        myProject = Project("My Project")
        projectRuns = myProject.runs()
        myRunId = projectRuns.query("title == 'myRun'").id.item()
        myRun = Run(myRunId)

    Attributes
    ----------
    id : str
        Project id
    name: str
        Project name
    attributes: dict
        Master attributes dictionary
    connection: transcriptic.config.Connection
        Transcriptic Connection object associated with this specific object

    """

    def __init__(self, project_id, attributes=None, connection=None):
        """
        Initialize a Project by providing a project name/id. The attributes and connection parameters are generally
        not specified unless one wants to manually initialize the object.

        Parameters
        ----------
        project_id: str
            Project name or id in string form
        attributes: Optional[dict]
            Attributes of the project
        connection: Optional[transcriptic.config.Connection]
            Connection context. The default context object will be used unless explicitly provided
        """
        super(Project, self).__init__("project", project_id, attributes, connection)
        self._runs = pd.DataFrame()

    def runs(self, use_cache=True):
        """
        Get the list of runs belonging to the project

        Parameters
        ----------
        use_cache: Boolean
            Determines whether the cached list of runs is returned

        Returns
        -------
        DataFrame
            Returns a DataFrame of runs, with the id and title as columns
        """
        if self._runs.empty and use_cache:
            temp = self.connection.env_args
            self.connection.update_environment(project_id=self.id)
            project_runs = self.connection.runs()
            self._runs = pd.DataFrame([[pr["id"], pr["title"]] for pr in project_runs])
            self._runs.columns = ["id", "Name"]
            self.connection.env_args = temp
        return self._runs

    def submit(self, protocol, title, test_mode=False):
        """
        Submit a run to this project

        Parameters
        ----------
        protocol: dict
            Autoprotocol Protocol in dictionary form, can be generated using Protocol.as_dict()
        title: Optional[str]
            Title of run. Run-id will automatically be used as name if field is not provided
        test_mode: Optional[boolean]
            Determines if run will be submitted will be treated as a test run or a run that is meant for execution

        Returns
        -------
        Run
            Returns a run object if run is successfully submitted
        """
        response = self.connection.submit_run(
            protocol, project_id=self.id, title=title, test_mode=test_mode
        )
        return Run(response["id"], response)


class Run(_BaseObject):
    """
    A Run object contains helper methods for accessing Run-related information such as Instructions, Datasets
    and monitoring data

    Example Usage:

        .. code-block:: python

            myRun = Run('r12345')
            myRun.data
            myRun.instructions
            myRun.containers
            myRun.Instructions[0]

    Attributes
    ----------
    id : str
        Run id
    name: str
        Run name
    data: DataFrame
        DataFrame summary of all datasets which belong to this run
    instructions: DataFrame
        DataFrame summary of all Instruction objects which belong to this run
    containers: DataFrame
        DataFrame summary of all Container objects which belong to this run
    project_id : str
        Project id which run belongs to
    attributes: dict
        Master attributes dictionary
    connection: transcriptic.config.Connection
        Transcriptic Connection object associated with this specific object

    """

    def __init__(self, run_id, attributes=None, connection=None, timeout=30.0):
        """
        Initialize a Run by providing a run name/id. The attributes and connection parameters are generally
        not specified unless one wants to manually initialize the object.

        Parameters
        ----------
        run_id: str
            Run name or id in string form
        attributes: Optional[dict]
            Attributes of the run
        connection: Optional[transcriptic.config.Connection]
            Connection context. The default context object will be used unless explicitly provided
        timeout: Optional[float]
            Timeout in seconds (defaults to 30.0). This will be used when making API calls to fetch data associated with the run.
        """
        super(Run, self).__init__("run", run_id, attributes, connection)
        self.project_id = self.attributes["project"]["id"]
        self.timeout = timeout
        self._data_ids = pd.DataFrame()
        self._instructions = pd.DataFrame()
        self._containers = pd.DataFrame()
        self._data = pd.DataFrame()

    @property
    def data_ids(self):
        """
        Find and generate a list of datarefs and data_ids associated with this run.

        Returns
        -------
        DataFrame
            Returns a DataFrame of data ids, with datarefs and data_ids as columns

        """
        if self._data_ids.empty:
            datasets = []
            for dataset in self.attributes["datasets"]:
                inst_id = dataset["instruction_id"]
                if inst_id:
                    titles = [
                        inst.attributes["operation"]["dataref"]
                        for inst in self.instructions["Instructions"]
                        if inst.attributes["id"] == inst_id
                    ]
                    if len(titles) == 0:
                        title = "unknown"
                    elif len(titles) == 1:
                        title = titles[0]
                    else:
                        # This should never happen since instruction_ids are unique
                        raise ValueError("No unique instruction id found")
                else:
                    title = dataset["title"]
                datasets.append(
                    {
                        "Name": title,
                        "DataType": dataset["data_type"],
                        "Id": dataset["id"],
                    }
                )
            if len(datasets) > 0:
                data_ids = pd.DataFrame(datasets)
                self._data_ids = data_ids[["Name", "DataType", "Id"]]
        return self._data_ids

    @property
    def instructions(self):
        if self._instructions.empty:
            instruction_list = [
                Instruction(
                    dict(x, **{"project_id": self.project_id, "run_id": self.id}),
                    connection=self.connection,
                )
                for x in self.attributes["instructions"]
            ]
            self._instructions = pd.DataFrame(instruction_list)
            self._instructions.columns = ["Instructions"]
            self._instructions.insert(
                0, "Name", [inst.name for inst in self._instructions.Instructions]
            )
            self._instructions.insert(
                1, "Id", [inst.id for inst in self._instructions.Instructions]
            )
            self._instructions.insert(
                2,
                "Started",
                [inst.started_at for inst in self._instructions.Instructions],
            )
            self._instructions.insert(
                3,
                "Completed",
                [inst.completed_at for inst in self._instructions.Instructions],
            )
        return self._instructions

    @property
    def Instructions(self):
        """
        Helper for allowing direct access of `Instruction` objects

        Returns
        -------
        Series
            Returns a Series of `Instruction` objects

        """
        return self.instructions.Instructions

    @property
    def containers(self):
        if self._containers.empty:
            container_list = []
            for ref in Run(self.id).attributes["refs"]:
                container_list.append(Container(ref["container"]["id"]))
            self._containers = pd.DataFrame(container_list)
            self._containers.columns = ["Containers"]
            self._containers.insert(
                0, "Name", [container.name for container in self._containers.Containers]
            )
            self._containers.insert(
                1,
                "ContainerId",
                [container.id for container in self._containers.Containers],
            )
            self._containers.insert(
                2,
                "Type",
                [
                    container.container_type.shortname
                    for container in self._containers.Containers
                ],
            )
            self._containers.insert(
                3,
                "Status",
                [
                    container.attributes["status"]
                    for container in self._containers.Containers
                ],
            )
            self._containers.insert(
                4,
                "Storage Condition",
                [container.storage for container in self._containers.Containers],
            )
        return self._containers

    @property
    def Containers(self):
        """
        Helper for allowing direct access of `Container` objects

        Returns
        -------
        Series
            Returns a Series of `Container` objects
        """
        return self.containers.Containers

    @property
    def data(self):
        """
        Find and generate a list of Dataset objects which are associated with this run

        Returns
        -------
        DataFrame
            Returns a DataFrame of datasets, with Name, Dataset and DataType as columns

        """
        if self._data.empty:
            num_datasets = len(self.data_ids)
            if num_datasets == 0:
                print("No datasets were found.")
            else:
                print("Attempting to fetch %d datasets..." % num_datasets)
                try:
                    data_list = []
                    for name, data_type, data_id in self.data_ids.values:
                        dataset = Dataset(data_id)
                        data_list.append(
                            {
                                "Name": name,
                                "DataType": data_type,
                                "Operation": dataset.operation,
                                "AnalysisTool": dataset.analysis_tool,
                                "Datasets": dataset,
                            }
                        )
                    data_frame = pd.DataFrame(data_list)

                    # Rearrange columns
                    self._data = data_frame[
                        ["Name", "DataType", "Operation", "AnalysisTool", "Datasets"]
                    ]
                except ReadTimeout:
                    print(
                        "Operation timed out after %d seconds. Returning data_ids instead of Datasets.\nTo try again, increase value of self.timeout and resubmit request."
                        % self.timeout
                    )
                    return self.data_ids
        return self._data

    @property
    def Datasets(self):
        """
        Helper for allowing direct access of `Dataset` objects

        Returns
        -------
        Series
            Returns a Series of `Dataset` objects
        """
        try:
            return self.data.Datasets
        except Exception:
            print("Unable to load Datasets successfully. Returning empty series.")
            return pd.Series()

    def _repr_html_(self):
        return """<iframe src="%s" frameborder="0" allowtransparency="true" \
        style="height:450px" seamless></iframe>""" % self.connection.get_route(
            "view_run", project_id=self.project_id, run_id=self.id
        )


class Dataset(_BaseObject):
    """
    A Dataset object contains helper methods for accessing data related information

    Attributes
    ----------
    id : str
        Dataset id
    name: str
        Dataset name
    data : DataFrame
        DataFrame of well-indexed data values. Note that associated metadata is found in attributes dictionary
    data_objects : list(DataObject)
        List of DataObject type
    attachments : dict(str, bytes)
        names and data of all attachments for the dataset
    container: Container
        Container object that was used for this dataset
    operation: str
        Operation used for generating the dataset
    data_type: str
        Data type of this dataset
    attributes: dict
        Master attributes dictionary
    connection: transcriptic.config.Connection
        Transcriptic Connection object associated with this specific object

    """

    def __init__(self, data_id, attributes=None, connection=None):
        """
        Initialize a Dataset by providing a data name/id. The attributes and connection parameters are generally
        not specified unless one wants to manually initialize the object.

        Parameters
        ----------
        data_id: str
            Dataset name or id in string form
        attributes: Optional[dict]
            Attributes of the dataset
        connection: Optional[transcriptic.config.Connection]
            Connection context. The default context object will be used unless explicitly provided
        """
        super(Dataset, self).__init__("dataset", data_id, attributes, connection)
        # TODO: Get BaseObject to handle dataset name
        self.name = self.attributes["title"]
        self.id = data_id

        # TODO: Consider more formally distinguishing between dataset types
        try:
            self.operation = self.attributes["instruction"]["operation"]["op"]
        except KeyError:
            self.operation = None
        try:
            self.container = Container(
                self.attributes["container"]["id"],
                attributes=self.attributes["container"],
                connection=connection,
            )
        except KeyError as e:
            if "instruction" in self.attributes:
                warnings.warn(f"Missing key {e} when initializing dataset")
            self.container = None

        self.analysis_tool = self.attributes["analysis_tool"]
        self.analysis_tool_version = self.attributes["analysis_tool_version"]
        self.data_type = self.attributes["data_type"]
        self._raw_data = None
        self._data = pd.DataFrame()
        self._attachments = None
        self._data_objects = None

    @property
    def attachments(self):
        if not self._attachments:
            self._attachments = self.connection.attachments(data_id=self.id)
        return self._attachments

    @property
    def raw_data(self):
        if not self._raw_data:
            # Get all raw data
            self._raw_data = self.connection.dataset(data_id=self.id, key="*")
        return self._raw_data

    @property
    def data(self, key="*"):
        if self._data.empty:
            # Get all data initially (think about lazy loading in the future)
            try:
                self._data = pd.DataFrame(self.raw_data)
            except:
                raise RuntimeError(
                    "Failed to cast data as DataFrame. Try using raw_data property instead."
                )
            self._data.columns = [x.upper() for x in self._data.columns]
        if key == "*":
            return self._data
        else:
            return self._data[key]

    def data_objects(self):
        if not self._data_objects:
            self._data_objects = DataObject.init_from_dataset_id(self.id)
        return self._data_objects

    def cross_ref_aliquots(self):
        # Use the container.aliquots DataFrame as the base
        aliquot_data = deepcopy(self.container.aliquots)
        data_column = []
        indices_without_data = []
        # Print a warning if new column will overwrite existing column
        if "Aliquot Data" in aliquot_data.columns.values.tolist():
            warnings.warn(
                "Column 'Aliquot Data' will be overwritten with data pulled from Dataset."
            )
        # Look up data for every well index
        for index in aliquot_data.index:
            # Get humanized index
            humanized_index = self.container.container_type.humanize(int(index))
            if humanized_index in self.data:
                # Use humanized index to get data for that well
                data_point = self.data.loc[0, humanized_index]
            else:
                # If no data for that well, use None instead
                data_point = None
                indices_without_data.append(humanized_index)
            # Append data point to list
            data_column.append(data_point)
        # Print a list of well indices that do not have corresponding data keys
        if len(indices_without_data) > 0:
            warnings.warn(
                "The following indices were not found as data keys: %s"
                % ", ".join(indices_without_data)
            )
        # Add these data as a column to the DataFrame
        aliquot_data["Aliquot Data"] = data_column

        return aliquot_data

    def _repr_html_(self):
        return """<iframe src="%s" frameborder="0" allowtransparency="true" \
            style="height:400px; width:600px" seamless></iframe>""" % self.connection.get_route(
            "view_data", data_id=self.id
        )


class DataObject(object):
    """
    A DataObject holds a reference to the raw data, stored in S3, along with format and validation information

    Attributes
    ----------
    id : str
        DataObject id
    dataset_id : str
        Dataset id
    data : bytes
        Bytes fetched from the url
    name: str
        Dataset name
    content_type: str
        content type
    format: str
        format
    size: int
        size in bytes
    status: Enum("valid", "invalid", "unverified")
        valid vs invalid
    url: str
        download url which expires every 1hr.  Call `refresh` to renew
    validation_errors: list(str)
        validation errors
    container: Container
        Container object that was used for this data object
    attributes: dict
        Master attributes dictionary
    """

    def __init__(self, data_object_id=None):
        attributes = {}

        # Fetch dataobject from server if id supplied
        if data_object_id is not None:
            attributes = DataObject.fetch_attributes(data_object_id)

        self.__init_attrs(attributes)

        # cached values
        self._container = None
        self._data = None
        self._json = None

    def __init_attrs(self, attributes):
        self.attributes = attributes

        self.id = attributes.get("id")
        self.dataset_id = attributes.get("dataset_id")
        self.content_type = attributes.get("content_type")
        self.format = attributes.get("format")
        self.name = attributes.get("name")
        self.size = attributes.get("size")
        self.status = attributes.get("status")
        self.url = attributes.get("url")
        self.validation_errors = attributes.get("validation_errors")

    @staticmethod
    def fetch_attributes(data_object_id):
        connection = _check_api("data_objects")
        return connection.data_object(data_object_id)

    @staticmethod
    def init_from_attributes(attributes):
        data_object = DataObject()
        data_object.__init_attrs(attributes)

        return data_object

    @staticmethod
    def init_from_id(data_object_id):
        return DataObject(data_object_id)

    @staticmethod
    def init_from_dataset_id(data_object_id):
        connection = _check_api("data_objects")

        # array of attributes
        attributes_arr = connection.data_objects(data_object_id)

        return [DataObject.init_from_attributes(a) for a in attributes_arr]

    @property
    def container(self):
        container_id = self.attributes["container_id"]

        if container_id is None:
            return None

        if not self._container:
            self._container = Container(container_id)

        return self._container

    @property
    def data(self):
        if self._data:
            return self._data

        self._data = requests.get(self.url).content

        return self._data

    @property
    def data_str(self):
        return self.data.decode("utf-8")

    @property
    def json(self):
        if self._json:
            return self._json

        self._json = json.loads(self.data)

        return self._json

    def dataframe(self):
        """Creates a simple Pandas Dataframe"""
        if self.format == "csv" or self.content_type == "text/csv":
            return pd.read_csv(StringIO(self.data_str))
        else:
            return pd.DataFrame(self.json)

    def save_data(self, filepath, chunk_size=1024):
        """Save DataObject data to a file.  Useful for large files"""
        with open(filepath, "wb") as f:
            if self._data:
                f.write(self._data)
                return

            r = requests.get(self.url, stream=True)

            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)

    def refresh(self):
        """Refresh DataObject as the url will expire after 1 hour"""
        clone = DataObject.init_from_id(self.id)
        self.__init_attrs(clone.attributes)


class Instruction(object):
    """
    An Instruction object contains information related to the current instruction such as the start,
    completed time as well as warps associated with the instruction.
    Note that Instruction objects are usually created as part of a run and not created explicity.

    Additionally, if diagnostic information is available, one can click on the `Show Diagnostics Data`
    button to view relevant diagnostic information.

    Example Usage:

        .. code-block:: python

            myRun = Run('r12345')
            myRun.instructions

            # Access instruction object
            myRun.Instructions[1]
            myRun.Instructions[1].warps


    Attributes
    ----------
    id : str
        Instruction id
    name: str
        Instruction name
    warps : DataFrame
        DataFrame of warps in the instruction
    started_at : str
        Time where instruction begun
    completed_at : str
        Time where instruction ended
    device_id: str
        Id of device which instruction was executed on
    attributes: dict
        Master attributes dictionary
    connection: transcriptic.config.Connection
        Transcriptic Connection object associated with this specific object
    """

    def __init__(self, attributes, connection=None):
        """
        Parameters
        ----------
        attributes : dict
            Instruction attributes
        connection: Optional[transcriptic.config.Connection]
            Connection context. The default context object will be used unless explicitly provided
        """
        self.connection = connection
        self.attributes = attributes
        self.id = attributes["id"]
        self.name = attributes["operation"]["op"]
        self.started_at = attributes["started_at"]
        self.completed_at = attributes["completed_at"]
        if len(attributes["warps"]) > 0:
            device_id_set = set(
                [warp["device_id"] for warp in self.attributes["warps"]]
            )
            self.device_id = device_id_set.pop()
            if len(device_id_set) > 1:
                warnings.warn(
                    "There is more than one device involved in this instruction. Please contact "
                    "Transcriptic for assistance."
                )

        else:
            self.device_id = None
        self._warps = pd.DataFrame()
        self._warp_events = pd.DataFrame()

    @property
    def warps(self):
        if self._warps.empty:
            warp_list = self.attributes["warps"]
            if len(warp_list) != 0:
                self._warps = pd.DataFrame(x["command"] for x in warp_list)
                self._warps.columns = [x.title() for x in self._warps.columns.tolist()]
                # Rearrange columns to start with `Name`
                if "Name" in self._warps.columns:
                    col_names = ["Name"] + [
                        col for col in self._warps.columns if col != "Name"
                    ]
                    self._warps = self._warps[col_names]
                self._warps.insert(1, "WarpId", [x["id"] for x in warp_list])
                self._warps.insert(
                    2, "Completed", [x["reported_completed_at"] for x in warp_list]
                )
                self._warps.insert(
                    3, "Started", [x["reported_started_at"] for x in warp_list]
                )
            else:
                warnings.warn(
                    "There are no warps associated with this instruction. Please contact "
                    "Transcriptic for assistance."
                )
        return self._warps

    @property
    def warp_events(self):
        """
        Warp events include discrete monitoring events such as liquid sensing
        events for a particular instruction.
        """
        # Note: We may consider adding special classes for specific warp
        # events, with more specific annotations/fields.
        if self._warp_events.empty:
            self._warp_events = self.monitoring(data_type="events")
        return self._warp_events

    def monitoring(self, data_type="pressure", grouping=None):
        """
        View monitoring data of a given instruction

        Parameters
        ----------
        data_type: Optional[str]
            Monitoring data type, defaults to 'pressure'
        grouping: Optional[str]
            Determines whether the values will be grouped, defaults to None. E.g. "5:ms"

        Returns
        -------
        DataFrame
            Returns a pandas dataframe of the monitoring data if present.
            Returns an empty dataframe if no data can be found due to errors.
        """
        response = self.connection.monitoring_data(
            instruction_id=self.id, data_type=data_type, grouping=grouping
        )
        # Handle errors by returning empty dataframe
        if "error" in response:
            warnings.warn(response["error"])
            return pd.DataFrame()
        res = pd.DataFrame(response["results"])
        # re-order so that "name" column is always leading
        if "name" in res.columns:
            rearr_cols = ["name"] + res.columns[res.columns != "name"].tolist()
            return res[rearr_cols]
        return res

    def _repr_html_(self):
        return """<iframe src="%s" frameborder="0" allowtransparency="true" \
            style="width:450px" seamless></iframe>""" % self.connection.get_route(
            "view_instruction",
            run_id=self.attributes["run_id"],
            project_id=self.attributes["project_id"],
            instruction_id=self.id,
        )


class Container(_BaseObject):
    """
    A Container object represents a container from the Transcriptic LIMS and
    contains relevant information on the container type as well as the
    aliquots present in the container.

    Example Usage:
        .. code-block:: python

          my_container = container("ct186apgz6a374")
          my_container.well_map
          my_container.aliquots

          my_container.container_type.col_count
          my_container.container_type.robotize("B1")
          my_container.container_type.humanize(12)

    Attributes
    ----------
    name: str
        Name of container
    well_map: dict
        Well mapping with well indices for keys and well names as values
    aliquots: DataFrame
        DataFrame of aliquots present in the container. DataFrame index
        now corresponds to the Well Index.
    container_type: autoprotocol.container_type.ContainerType
        Autoprotocol ContainerType object with many useful container type
        information and functions.
    cover: str
        Cover type of container
    storage: str
        Storage condition of container

        Example Usage:

        .. code-block:: python

          my_container = container("ct186apgz6a374")

          my_container.well_map

          my_container.container_type.col_count
          my_container.container_type.robotize("B1")
          my_container.container_type.humanize(12)


    """

    def __init__(self, container_id, attributes=None, connection=None):
        """
        Initialize a Container by providing a container name/id. The attributes and connection parameters are generally
        not specified unless one wants to manually initialize the object.

        Parameters
        ----------
        container_id: str
            Container name or id in string form
        attributes: Optional[dict]
            Attributes of the container
        connection: Optional[transcriptic.config.Connection]
            Connection context. The default context object will be used unless explicitly provided
        """
        super(Container, self).__init__(
            "container", container_id, attributes, connection
        )
        # TODO: Unify container "label" with name, add Containers route
        self.id = container_id
        self.cover = self.attributes["cover"]
        self.name = self.attributes["label"]
        self.storage = self.attributes["storage_condition"]
        self.well_map = {
            aliquot["well_idx"]: aliquot["name"]
            for aliquot in self.attributes["aliquots"]
        }
        self.container_type = self._parse_container_type()
        self._aliquots = pd.DataFrame()

    def _parse_container_type(self):
        """Helper function for parsing container string into container object"""

        container_type = self.attributes["container_type"]

        # Return the corresponding AP-Py container object for now. In the future, consider merging
        # the current and future dictionary when instantiating container_type
        try:
            from autoprotocol.container_type import _CONTAINER_TYPES

            return _CONTAINER_TYPES[container_type["shortname"]]
        except ImportError:
            warnings.warn(
                "Please install `autoprotocol-python` in order to get container types"
            )
            return None
        except KeyError:
            warnings.warn("ContainerType given is not supported yet in AP-Py")
            return None

    @property
    def aliquots(self):
        """
        Return a DataFrame of aliquots in the container, along with aliquot
        name, volume, and properties. Row index for the DataFrame corresponds
        to the well index of the aliquot.

        """
        if self._aliquots.empty:
            aliquot_list = self.attributes["aliquots"]
            try:
                from autoprotocol import Unit

                self._aliquots = pd.DataFrame(
                    sorted(
                        [
                            dict(
                                {
                                    "Well Index": x["well_idx"],
                                    "Name": x["name"],
                                    "Id": x["id"],
                                    "Volume": Unit(float(x["volume_ul"]), "microliter"),
                                },
                                **x["properties"],
                            )
                            for x in aliquot_list
                        ],
                        key=itemgetter("Well Index"),
                    )
                )
            except ImportError:
                warnings.warn(
                    "Volume is not cast into Unit-type. Please install `autoprotocol-python` in order to have automatic Unit casting"
                )
                self._aliquots = pd.DataFrame(
                    sorted(
                        [
                            dict(
                                {
                                    "Well Index": x["well_idx"],
                                    "Name": x["name"],
                                    "Id": x["id"],
                                    "Volume": float(x["volume_ul"]),
                                },
                                **x["properties"],
                            )
                            for x in aliquot_list
                        ],
                        key=itemgetter("Well Index"),
                    )
                )
            indices = self._aliquots.pop("Well Index")
            self._aliquots.set_index(indices, inplace=True)
        return self._aliquots

    def __repr__(self):
        """
        Return a string representation of a Container using the specified name.
        (ex. Container('my_plate'))

        """
        return "Container(%s)" % (str(self.name))



from .config import Connection
from .version import __version__

api = None

"""
Transcriptic
============

The Transcriptic library is separated into three components:
1) Core. The core modules provide a barebones client for making calls to
the Transcriptic webapp to create and obtain data. This can be done via the
`api` object or via the command-line using the CLI.
2) Jupyter. This module provides a Jupyter-centric means for interacting with
objects returned from the Transcriptic webapp such as Run, Project and Dataset.
3) Analysis. This module provides some basic analysis wrappers around datasets
returned from the webapp using standard Python scientific libraries.

The __init__ file contains a bunch of entry functions to facilitate easy access
to the Jupyter library.
"""

""" Allow direct import of Jupyter objects such as `Run` directly if the
required imports are present
"""
try:
    from .jupyter import Run, Project, Container, Dataset
    from .commands import ProtocolPreview
except ImportError as e:
    pass


def run(run_id):
    """
    Creates a Run object for the specified run-id

    Parameters
    ----------
    run_id: str
        Id of Run, e.g. r123456789
    Returns
    ------
    Run object: Run
        Transcriptic representation of a Run object
    """
    from .jupyter import Run

    return Run(run_id)


def project(project_id):
    """
    Creates a Project object for the specified project-id

    Parameters
    ----------
    project_id: str
        Id of Project, e.g. p123456789
    Returns
    ------
    Project object: Project
        Transcriptic representation of a Project object
    """
    from .jupyter import Project

    return Project(project_id)


def container(container_id):
    """
    Creates a Container object for the specified container-id

    Parameters
    ----------
    container_id: str
        Id of Container, e.g. ct123456789
    Returns
    ------
    Container object: Container
        Transcriptic representation of a Container object
    """
    from .jupyter import Container

    return Container(container_id)


def preview(protocol):
    """
    Creates a protocol preview object for the specified protocol

    Parameters
    ----------
    protocol: dictionary
        protocol JSON in dictionary format
    Returns
    ------
    Protocol object: ProtocolPreview
        Transcriptic representation of a Protocol object
    """
    from .commands import ProtocolPreview

    return ProtocolPreview(protocol, api=api)


def analyze(protocol, test_mode=False):
    """
    Analyze a given protocol

    Parameters
    ----------
    protocol: dict
        Autoprotocol JSON in dictionary format
    test_mode: boolean
        whether protocol should be analyzed under test mode
    Returns
    ------
    Analysis result: dict
        Raw result of the analysis
    """
    return api.analyze_run(protocol, test_mode)


def submit(protocol, project_id, title=None, test_mode=False):
    """
    Submit a given protocol

    Parameters
    ----------
    protocol: dict
        Autoprotocol JSON in dictionary format
    project_id: str
        Project to submit Autoprotocol to
    title: str
        Name of Run
    test_mode: boolean
        whether protocol should be submitted as a test run
    Returns
    ------
    Submission result: dict
        Raw result of the submission
    """
    return api.submit_run(
        protocol, project_id=project_id, title=title, test_mode=test_mode
    )


def dataset(data_id, key="*"):
    """
    Get dataset for a given data_id

    Parameters
    ----------
    data_id
        Id of desired dataset, e.g. d123456789
    key
        Key of desired sub-fields of dataset

    Returns
    ------
    Data: dict
        Data in JSON form
    """
    return api.dataset(data_id=data_id, key=key)


def connect(transcriptic_path="~/.transcriptic"):
    """
    Instantiates a Connection based on the specified path, and overwrites the
    existing `api` object with this Connection

    Parameters
    ----------
    transcriptic_path:
     Path to transcriptic dot-file
    """
    # TODO: Mirror login code from CLI
    try:
        api = Connection.from_file(transcriptic_path)
    except (OSError, IOError):
        print(
            "Unable to find .transcriptic file, please ensure the right path"
            " is provided"
        )



import inspect
import io
import json
import os
import platform
import requests
import time
from Crypto.PublicKey import RSA
import transcriptic
import warnings
import zipfile

from . import routes
from .signing import StrateosSign, BearerAuth
from .util import is_valid_jwt_token
from .version import __version__

try:
    import magic
except ImportError:
    warnings.warn(
        "`python-magic` is recommended. You may be missing some system-level "
        "dependencies if you have already pip-installed it.\n"
        "Please refer to https://github.com/ahupp/python-magic#installation "
        "for more installation instructions."
    )


def initialize_default_session():
    """
    Initialize a default `requests.Session()` object which can be used for
    requests into the tx web api.
    """
    session = requests.Session()
    session.headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "User-Agent": f"txpy/{__version__} "
        f"({platform.python_implementation()}/"
        f"{platform.python_version()}; "
        f"{platform.system()}/{platform.release()}; "
        f"{platform.machine()}; {platform.architecture()[0]})",
    }
    return session


class Connection(object):
    """
    A Connection object is the object used for communicating with Transcriptic.

    Local usage: This is most easily instantiated by using the `from_file`
    function after calling `transcriptic login` from the command line.

    .. code-block:: shell
        :caption: shell

        $ transcriptic login
        Email: me@example.com
        Password:
        Logged in as me@example.com (example-lab)

    .. code-block:: python
        :caption: python

        from transcriptic.config import Connection
        api = Connection.from_file("~/.transcriptic")

    For those using Jupyter notebooks on secure.transcriptic.com (beta), a
    Connection object is automatically instantiated as api.

    .. code-block:: python
        :caption: python

        from transcriptic import api

    The `api` object can then be used for making any api calls. It is
    recommended to use the objects in `transcriptic.objects` since that wraps
    the response in a more friendly format.

    Example Usage:

    .. code-block:: python
        :caption: python

        api.projects()
        api.runs(project_id="p123456789")

    If you have multiple organizations and would like to switch to a specific
    organization, or if you would like to auto-load certain projects, you can
    set it directly by assigning to the corresponding variable.

    Example Usage:

    .. code-block:: python
        :caption: python

        api.organization_id = "my_other_org"
        api.project_id = "p123"

    """

    def __init__(
        self,
        email=None,
        token=None,
        organization_id=None,
        api_root="https://secure.transcriptic.com",
        cookie=None,
        verbose=False,
        analytics=True,
        user_id="default",
        feature_groups=[],
        rsa_key=None,
        session=None,
        bearer_token=None,
    ):
        # Initialize environment args used for computing routes
        self.env_args = dict()
        self.api_root = api_root
        self.organization_id = organization_id

        # Initialize session headers
        if session is None:
            session = initialize_default_session()
        self.session = session

        self._bearer_token = None

        # Initialize RSA props
        self._rsa_key = None
        self._rsa_key_path = None
        self._rsa_secret = None
        # Set/Load rsa_key from arguement as string or Path
        self.rsa_key = rsa_key

        # NB: These many setattr calls update self.session.headers
        # cookie authentication is mutually exclusive from token authentication
        if cookie:
            if email is not None or token is not None:
                warnings.warn(
                    "Cookie and token authentication is mutually "
                    "exclusive. Ignoring email and token"
                )
            self.session.headers["X-User-Email"] = None
            self.session.headers["X-User-Token"] = None
            self.cookie = cookie
            self.update_session_auth(use_signature=False)
        else:
            if cookie is not None:
                warnings.warn(
                    "Cookie and token authentication is mutually "
                    "exclusive. Ignoring cookie"
                )
            self.session.headers["Cookie"] = None
            self.email = email
            if token is not None:
                self.token = token
                if bearer_token is not None:
                    warnings.warn(
                        "User token and bearer token authentication"
                        "is mutually exclusive. Ignoring bearer token"
                    )
            elif bearer_token is not None:
                self.bearer_token = bearer_token
            self.update_session_auth()

        # Initialize feature groups
        self.feature_groups = feature_groups

        # Initialize CLI parameters
        self.verbose = verbose
        self.analytics = analytics
        self.user_id = user_id

        transcriptic.api = self

    @staticmethod
    def from_file(path):
        """Loads connection from file"""
        config_path = os.path.expanduser(path)
        with open(config_path) as f:
            cfg = json.load(f)

        expected_keys = set(
            ("email", "token", "organization_id", "api_root", "analytics", "user_id")
        )
        keys = set(cfg.keys())

        if not keys.issuperset(expected_keys):
            raise OSError(
                f"Key(s) not found in configuration file ({config_path}) "
                f"Missing {repr(expected_keys - keys)}"
            )
        return Connection(**cfg)

    @staticmethod
    def from_default_config():
        """
        Load the default configuration file from the home directory of the
        current user and return a Connection instance that is costructed from it.
        """
        return Connection.from_file("~/.transcriptic")

    @property
    def api_root(self):
        try:
            return self.env_args["api_root"]
        except (NameError, KeyError):
            raise ValueError("api_root is not set.")

    @api_root.setter
    def api_root(self, value):
        self.update_environment(api_root=value)

    @property
    def organization_id(self):
        try:
            return self.env_args["org_id"]
        except (NameError, KeyError):
            raise ValueError("organization_id is not set.")

    @organization_id.setter
    def organization_id(self, value):
        self.update_environment(org_id=value)

    @property
    def project_id(self):
        try:
            return self.env_args["project_id"]
        except (NameError, KeyError):
            raise ValueError("project_id is not set.")

    @project_id.setter
    def project_id(self, value):
        self.update_environment(project_id=value)

    @property
    def email(self):
        try:
            return self.session.headers["X-User-Email"]
        except (NameError, KeyError):
            raise ValueError("email is not set.")

    @email.setter
    def email(self, value):
        if self.cookie is not None:
            warnings.warn(
                "Cookie and token authentication is mutually "
                "exclusive. Clearing cookie from headers"
            )
            self.update_headers(**{"Cookie": None})
        self.update_headers(**{"X-User-Email": value})
        self.update_session_auth()

    @property
    def bearer_token(self):
        return self._bearer_token

    @bearer_token.setter
    def bearer_token(self, value):
        if is_valid_jwt_token(value):
            self._bearer_token = value
        else:
            raise ValueError("Malformed JWT Bearer Token")

    @property
    def token(self):
        try:
            return self.session.headers["X-User-Token"]
        except (NameError, KeyError):
            raise ValueError("token is not set.")

    @token.setter
    def token(self, value):
        if self.cookie is not None:
            warnings.warn(
                "Cookie and token authentication is mutually "
                "exclusive. Clearing cookie from headers"
            )
            self.update_headers(**{"Cookie": None})
        self.update_headers(**{"X-User-Token": value})

    @property
    def cookie(self):
        try:
            return self.session.headers["Cookie"]
        except (NameError, KeyError):
            return ValueError("cookie is not set.")

    @cookie.setter
    def cookie(self, value):
        if self.email is not None or self.token is not None:
            warnings.warn(
                "Cookie and token authentication is mutually "
                "exclusive. Clearing email and token from headers"
            )
            self.update_headers(**{"X-User-Email": None, "X-User-Token": None})
        self.update_headers(**{"Cookie": value})

    @property
    def rsa_key(self):
        return self._rsa_key

    @rsa_key.setter
    def rsa_key(self, value):
        self._rsa_key = value
        self.load_rsa_secret()

    def save(self, path):
        """Saves current connection into specified file, used for CLI"""
        with open(os.path.expanduser(path), "w") as f:
            f.write(
                json.dumps(
                    {
                        "email": self.email,
                        "token": self.token,
                        "organization_id": self.organization_id,
                        "api_root": self.api_root,
                        "analytics": self.analytics,
                        "user_id": self.user_id,
                        "feature_groups": self.feature_groups,
                        # We dont want to save a string key, only a path
                        "rsa_key": self._rsa_key_path,
                    },
                    indent=2,
                )
            )

    def load_rsa_secret(self):
        key_string_or_path = self._rsa_key

        if key_string_or_path is None:
            self._rsa_key_path = None
            self._rsa_secret = None
        else:
            try:  # First try loading it as a key
                key = RSA.import_key(key_string_or_path)
                self._rsa_key_path = None
                self._rsa_secret = key.export_key()

            except ValueError:  # Then try as a Path
                key_path = os.path.abspath(os.path.expanduser(key_string_or_path))
                with open(key_path, "rb") as key_file:
                    key = RSA.import_key(key_file.read())
                self._rsa_key_path = str(key_path)
                self._rsa_secret = key.export_key()

        self.update_session_auth()

    def update_session_auth(self, use_signature=True):
        if (
            use_signature
            and self._rsa_secret
            and "X-User-Email" in self.session.headers
        ):
            self.session.auth = StrateosSign(self.email, self._rsa_secret)
        elif self.bearer_token:
            self.session.auth = BearerAuth(self.bearer_token)
        else:
            self.session.auth = None

    def update_environment(self, **kwargs):
        """
        Updates environment variables used for computing routes.
        To remove an existing variable, set value to None.
        """
        self.env_args = dict(self.env_args, **kwargs)

    def update_headers(self, **kwargs):
        """
        Updates session headers
        To remove an existing variable, set value to None.
        """
        self.session.headers = dict(self.session.headers, **kwargs)

    def url(self, path):
        """url format helper"""
        if path.startswith("/"):
            return f"{self.api_root}{path}"
        else:
            return f"{self.api_root}/{self.organization_id}/{path}"

    def preview_protocol(self, protocol):
        """Post protocol preview"""
        route = self.get_route("preview_protocol")
        protocol = _parse_protocol(protocol)
        err_default = "Unable to preview protocol"
        return self.post(
            route,
            json={"protocol": protocol},
            allow_redirects=False,
            status_response={
                "200": lambda resp: resp.json()["key"],
                "default": lambda resp: RuntimeError(err_default),
            },
        )

    def organizations(self):
        """Get list of organizations"""
        route = self.get_route("get_organizations")
        return self.get(route)

    def get_organization(self, org_id=None):
        """Get particular organization"""
        route = self.get_route("get_organization", org_id=org_id)
        err_404 = f"There was an error fetching the organization {org_id}"
        resp = self.get(
            route,
            status_response={
                "200": lambda resp: resp,
                "404": lambda resp: RuntimeError(err_404),
                "default": lambda resp: resp,
            },
        )
        return resp

    def projects(self):
        """Get list of projects in organization"""
        route = self.get_route("get_projects")
        err_default = (
            "There was an error listing the projects in your "
            "organization.  Make sure your login details are correct."
        )
        return self.get(
            route,
            status_response={
                "200": lambda resp: resp.json()["projects"],
                "default": lambda resp: RuntimeError(err_default),
            },
        )

    def project(self, project_id=None):
        """Get particular project"""
        route = self.get_route("get_project", project_id=project_id)
        err_default = f"There was an error fetching project {project_id}"
        return self.get(
            route, status_response={"default": lambda resp: RuntimeError(err_default)}
        )

    def runs(self, project_id=None):
        """Get list of runs in project"""
        route = self.get_route("get_project_runs", project_id=project_id)
        err_default = (
            f"There was an error fetching the runs in project " f"{project_id}"
        )
        return self.get(
            route,
            status_response={
                "200": lambda resp: resp.json(),
                "default": lambda resp: RuntimeError(err_default),
            },
        )

    def create_project(self, title):
        """Create project with given title"""
        route = self.get_route("create_project")
        return self.post(route, data=json.dumps({"name": title}))

    def delete_project(self, project_id=None):
        """Delete project with given project_id"""
        route = self.get_route("delete_project", project_id=project_id)
        return self.delete(route, status_response={"200": lambda resp: True})

    def archive_project(self, project_id=None):
        """Archive project with given project_id"""
        route = self.get_route("archive_project", project_id=project_id)
        return self.put(
            route,
            data=json.dumps({"project": {"archived": True}}),
            status_response={"200": lambda resp: True},
        )

    def modify_aliquot_properties(
        self, aliquot_id, set_properties={}, delete_properties=[]
    ):
        """
        Modify the properties of an alquot:

        If specified set_properties must be dict mapping {str: str}
        these properties will be set on the aliquot specified.

        If specified delete_properties must be a list of string
        properties which will be deleted on the aliquot.
        """
        route = self.get_route("modify_aliquot_properties", aliquot_id=aliquot_id)
        return self.put(
            route,
            json={
                "id": aliquot_id,
                "data": {"set": set_properties, "delete": delete_properties},
            },
        )

    def packages(self):
        """Get list of packages in organization"""
        route = self.get_route("get_packages")
        return self.get(route)

    def package(self, package_id=None):
        """Get package with given package_id"""
        route = self.get_route("get_package", package_id=package_id)
        return self.get(route)

    def create_package(self, name, description):
        """Create package with given name and description"""
        route = self.get_route("create_package")
        return self.post(
            route,
            data=json.dumps(
                {
                    "name": "%s%s" % ("com.%s." % self.organization_id, name),
                    "description": description,
                }
            ),
        )

    def delete_package(self, package_id=None):
        """Delete package with given package_id"""
        route = self.get_route("delete_package", package_id=package_id)
        return self.delete(route, status_response={"200": lambda resp: True})

    def post_release(self, data, package_id=None):
        """Create release with given data and package_id"""
        route = self.get_route("post_release", package_id=package_id)
        return self.post(route, data=data)

    def get_release_status(self, package_id=None, release_id=None, timestamp=None):
        """Get status of current release upload"""
        route = self.get_route(
            "get_release_status",
            package_id=package_id,
            release_id=release_id,
            timestamp=timestamp,
        )
        return self.get(route)

    def get_quick_launch(self, project_id=None, quick_launch_id=None):
        """Get quick launch object"""
        route = self.get_route(
            "get_quick_launch", project_id=project_id, quick_launch_id=quick_launch_id
        )
        return self.get(route)

    def create_quick_launch(self, data, project_id=None):
        """Create quick launch object"""
        route = self.get_route("create_quick_launch", project_id=project_id)
        return self.post(route, data=data)

    def launch_protocol(self, params, protocol_id=None):
        """Launch protocol-id with params"""
        route = self.get_route("launch_protocol", protocol_id=protocol_id)
        return self.post(route, data=params)

    def get_launch_request(self, protocol_id=None, launch_request_id=None):
        """Get launch request id"""
        route = self.get_route(
            "get_launch_request",
            protocol_id=protocol_id,
            launch_request_id=launch_request_id,
        )
        return self.get(route)

    def resolve_quick_launch_inputs(
        self, raw_inputs, project_id=None, quick_launch_id=None
    ):
        """Resolves `raw_inputs` to `inputs` for quick_launch"""
        route = self.get_route(
            "resolve_quick_launch_inputs",
            project_id=project_id,
            quick_launch_id=quick_launch_id,
        )
        return self.post(route, json=raw_inputs)

    def get_protocols(self):
        """Get list of available protocols"""
        route = self.get_route("get_protocols")
        return self.get(route)

    def resources(self, query):
        """Get resources"""
        route = self.get_route("query_resources", query=query)
        return self.get(route)

    def inventory(self, query, timeout=30.0, page=0):
        """Get inventory"""
        route = self.get_route("query_inventory", query=query, page=page)
        return self.get(route, timeout=timeout)

    def kits(self, query):
        """Get kits"""
        route = self.get_route("query_kits", query=query)
        return self.get(route)

    def payment_methods(self):
        route = self.get_route("get_payment_methods")
        return self.get(route)

    def monitoring_data(
        self, data_type, instruction_id, grouping=None, start_time=None, end_time=None
    ):
        """Get monitoring_data"""
        route = self.get_route(
            "monitoring_data",
            data_type=data_type,
            instruction_id=instruction_id,
            grouping=grouping,
            start_time=start_time,
            end_time=end_time,
        )
        return self.get(route)

    def raw_image_data(self, data_id=None):
        """Get raw image data"""
        route = self.get_route("view_raw_image", data_id=data_id)
        return self.get(route, status_response={"200": lambda resp: resp}, stream=True)

    def _get_object(self, obj_id, obj_type=None):
        """Helper function for loading objects"""
        # TODO: Migrate away from deref routes for other object types
        if obj_type == "dataset":
            route = self.get_route("dataset_short", data_id=obj_id)
        else:
            route = self.get_route("deref_route", obj_id=obj_id)
        err_404 = f"[404] No object found for ID {obj_id}"
        return self.get(route, status_response={"404": lambda resp: Exception(err_404)})

    def analyze_run(self, protocol, test_mode=False):
        """Analyze given protocol"""
        protocol = _parse_protocol(protocol)
        if "errors" in protocol:
            raise AnalysisException(
                (
                    "Error%s in protocol:\n%s"
                    % (
                        ("s" if len(protocol["errors"]) > 1 else ""),
                        "".join(
                            ["- " + e["message"] + "\n" for e in protocol["errors"]]
                        ),
                    )
                )
            )

        def error_string(r):
            return AnalysisException(
                "Error%s in protocol:\n%s"
                % (
                    ("s" if len(r.json()["protocol"]) > 1 else ""),
                    "".join(["- " + e["message"] + "\n" for e in r.json()["protocol"]]),
                )
            )

        return self.post(
            self.get_route("analyze_run"),
            data=json.dumps({"protocol": protocol, "test_mode": test_mode}),
            status_response={"422": lambda resp: error_string(resp)},
        )

    def submit_run(
        self,
        protocol,
        project_id=None,
        title=None,
        test_mode=False,
        payment_method_id=None,
    ):
        """Submit given protocol"""
        protocol = _parse_protocol(protocol)
        payload = {
            "title": title,
            "protocol": protocol,
            "test_mode": test_mode,
            "payment_method_id": payment_method_id,
        }
        data = {k: v for k, v in payload.items() if v is not None}
        route = self.get_route("submit_run", project_id=project_id)
        err_404 = (
            f"Error: Couldn't create run (404).\n Are you sure the "
            f"project {self.url(project_id)} exists, and that you have "
            f"access to it?"
        )

        def err_422(resp):
            f"Error creating run: {resp.text}"

        return self.post(
            route,
            data=json.dumps(data),
            status_response={
                "404": lambda resp: AnalysisException(err_404),
                "422": lambda resp: AnalysisException(err_422),
            },
        )

    def analyze_launch_request(self, launch_request_id, test_mode=False):
        return self.post(
            self.get_route("analyze_launch_request"),
            data=json.dumps(
                {"launch_request_id": launch_request_id, "test_mode": test_mode}
            ),
        )

    def submit_launch_request(
        self,
        launch_request_id,
        project_id=None,
        protocol_id=None,
        title=None,
        test_mode=False,
        payment_method_id=None,
    ):
        """Submit specified launch request"""
        payload = {
            "title": title,
            "launch_request_id": launch_request_id,
            "protocol_id": protocol_id,
            "test_mode": test_mode,
            "payment_method_id": payment_method_id,
        }
        data = {k: v for k, v in payload.items() if v is not None}
        return self.post(
            self.get_route("submit_launch_request", project_id=project_id),
            data=json.dumps(data),
            status_response={
                "404": lambda resp: AnalysisException(
                    "Error: Couldn't create run (404). \n"
                    "Are you sure the project %s "
                    "exists, and that you have access to it?" % self.url(project_id)
                ),
                "422": lambda resp: AnalysisException(
                    f"Error creating run: {resp.text}"
                ),
            },
        )

    def dataset(self, data_id, key="*"):
        """Get dataset with given data_id"""
        route = self.get_route("dataset", data_id=data_id, key=key)
        return self.get(route)

    def _get_uploads_from_key(self, key):
        """Fetches uploads for a data upload key

        Parameters
        ----------
        key : str
            data upload key

        Returns
        -------
        requests.Response
        """
        return self.get(
            route=self.get_route(method="get_uploads", key=key),
            status_response={"200": lambda resp: resp},
            stream=True,
        )

    def attachments(self, data_id):
        """Fetches all attachments for a given dataset id

        Parameters
        ----------
        data_id : str
            dataset id

        Returns
        -------
        dict(str, bytes)
            a dict of attachment names and contents

        """
        dataset_route = self.get_route("dataset_short", data_id=data_id)
        dataset_attachments = self.get(dataset_route).get("attachments")
        attachment_names = [
            os.path.basename(_.get("name", "")) for _ in dataset_attachments
        ]
        attachment_contents = [
            self._get_uploads_from_key(_.get("key")).content
            for _ in dataset_attachments
        ]
        return dict(zip(attachment_names, attachment_contents))

    def datasets(self, project_id=None, run_id=None, timeout=30.0):
        """Get datasets belonging to run"""
        route = self.get_route("datasets", project_id=project_id, run_id=run_id)
        err_404 = (
            f"[404] No run found for ID {run_id}. Please ensure you "
            f"have the right permissions."
        )
        return self.get(
            route,
            status_response={"404": lambda resp: Exception(err_404)},
            timeout=timeout,
        )

    def data_object(self, id):
        """Fetches a data object by id

        Parameters
        ----------
        id: str
            data_object id

        Returns
        -------
        dict
            attributes dict
        """
        route = self.get_route("data_object", id=id)
        attributes = self.get(route).get("data").get("attributes")
        attributes["id"] = id

        return attributes

    def data_objects(self, dataset_id):
        """Fetches all data objects given a dataset id

        Parameters
        ----------
        dataset_id: str
            dataset id

        Returns
        -------
        arr[dict]
            array of attributes dict
        """
        route_base = self.get_route("data_objects", dataset_id=dataset_id)
        page = 0
        limit = 50
        has_more = True

        results = []

        while has_more:
            route = f"{route_base}&page[limit]={limit}&page[offset]=" f"{page * limit}"
            response = self.get(route)
            entities = response.get("data")

            for entity in entities:
                attributes = entity["attributes"]
                attributes["id"] = entity["id"]
                results.append(attributes)

            page += 1

            if len(entities) != limit:
                has_more = False

        return results

    def upload_dataset_from_filepath(
        self, file_path, title, run_id, analysis_tool, analysis_tool_version
    ):
        """
        Helper for uploading a file as a dataset to the specified run.

        Uses `upload_dataset`.

        .. code-block:: python

            api.upload_dataset_from_filepath(
                "my_file.txt",
                title="my cool dataset",
                run_id="r123",
                analysis_tool="cool script",
                analysis_tool_version="v1.0.0"
            )

        Parameters
        ----------
        file: str
            Path to file to be uploaded
        title: str
            Name of dataset
        run_id: str
            Run-id
        analysis_tool: str, optional
            Name of tool used for analysis
        analysis_tool_version: str, optional
            Version of tool used

        Returns
        -------
        response: dict
            JSON-formatted response
        """
        try:
            file_path = os.path.expanduser(file_path)
            file_handle = open(file_path, "rb")
            name = os.path.basename(file_handle.name)
        except (AttributeError, FileNotFoundError):
            raise ValueError("'file' has to be a valid filepath")

        try:
            content_type = magic.from_file(file_path, mime=True)
        except NameError:
            # Handle issues with magic import by not decoding content_type
            content_type = None
        return self.upload_dataset(
            file_handle,
            name,
            title,
            run_id,
            analysis_tool,
            analysis_tool_version,
            content_type,
        )

    def upload_dataset(
        self,
        file_handle,
        name,
        title,
        run_id,
        analysis_tool,
        analysis_tool_version,
        content_type=None,
    ):
        """
        Uploads a file_handle as a dataset to the specified run.

        .. code-block:: python

            # Uploading a data_frame via file_handle, using Py3
            import io

            temp_buffer = io.StringIO()
            my_df.to_csv(temp_buffer)

            api.upload_dataset(
                temp_buffer,
                name="my_df",
                title="my cool dataset",
                run_id="r123",
                analysis_tool="cool script",
                analysis_tool_version="v1.0.0"
            )

        Parameters
        ----------
        file_handle: file_handle
            File handle to be uploaded
        name: str
            Dataset filename
        title: str
            Name of dataset
        run_id: str
            Run-id
        analysis_tool: str, optional
            Name of tool used for analysis
        analysis_tool_version: str, optional
            Version of tool used
        content_type: str
            Type of content uploaded

        Returns
        -------
        response: dict
            JSON-formatted response
        """
        upload_id = self.upload_to_uri(file_handle, content_type, title, name)
        upload_datasets_route = self.get_route("upload_datasets")
        upload_resp = self.post(
            upload_datasets_route,
            json={
                "upload_id": upload_id,
                "file_name": name,
                "title": title,
                "run_id": run_id,
                "analysis_tool": analysis_tool,
                "analysis_tool_version": analysis_tool_version,
            },
            status_response={
                "404": lambda resp: "[404] Please double-check your parameters"
                " and ensure they are valid."
            },
        )

        return upload_resp

    def upload_to_uri(self, file_handle, content_type, title, name):
        """
        Helper for uploading files via the `upload` route

        Parameters
        ----------
        file_handle: file_handle
            File handle to be uploaded
        content_type: str
            Type of content uploaded
        title: str
            Title of content to be uploaded
        name: str
            Name of file to be uploaded

        Returns
        -------
        key: str
            s3 key
        """
        # NOTE(meawoppl) title argument is unused?

        # TODO:
        # Currently, we are passing `0` for file_size as it doesn't really
        # matter for non multipart uploads, though it would be better to
        # supply the correct value.
        data = {
            "attributes": {
                "file_name": name,
                "file_size": 0,
                "last_modified": int(time.time()),
                "is_multipart": False,
            }
        }

        uri_route = self.get_route("upload")
        uri_resp = self.post(uri_route, data=json.dumps({"data": data}))

        try:
            upload_id = uri_resp["data"]["id"]
            upload_uri = uri_resp["data"]["attributes"]["upload_url"]
        except KeyError:
            raise RuntimeError("Unexpected payload returned for upload_dataset")

        if isinstance(file_handle, io.StringIO):
            try:
                # io.StringIO instances must be converted to bytes
                file_handle = io.BytesIO(bytes(file_handle.getvalue(), "utf-8"))
            except AttributeError:
                raise ValueError("Unable to convert read buffer to bytes")

        headers = {
            "Content-Disposition": f"attachment; filename={name}",
            "Content-Type": content_type,
        }
        headers = {k: v for k, v in headers.items() if v}
        self.put(
            upload_uri,
            data=file_handle,
            headers=headers,
            status_response={"200": lambda resp: resp},
        )
        return upload_id

    def get_zip(self, data_id, file_path=None):
        """
        Get zip file with given data_id. Downloads to memory and returns a
        Python ZipFile by default.
        When dealing with larger files where it may not be desired to load the
        entire file into memory, specifying `file_path` will enable the file to
        be downloaded locally.

        Example Usage:

        .. code-block:: python

            small_zip_id = 'd12345'
            small_zip = api.get_zip(small_zip_id)

            my_big_zip_id = 'd99999'
            api.get_zip(my_big_zip_id, file_path='big_file.zip')

        Parameters
        ----------
        data_id: data_id
            Data id of file to download
        file_path: Optional[str]
            Path to file which to save the response to. If specified, will not
            return ZipFile explicitly.

        Returns
        ----------
        zip: zipfile.ZipFile
            A Python ZipFile is returned unless `file_path` is specified

        """
        route = self.get_route("get_data_zip", data_id=data_id)
        req = self.get(route, status_response={"200": lambda resp: resp}, stream=True)

        if file_path:
            with open(file_path, "wb") as f:
                # Buffer download of data into memory with smaller chunk sizes
                chunk_sz = 1024  # 1kb chunks
                for chunk in req.iter_content(chunk_sz):
                    if chunk:
                        f.write(chunk)
            print(f"Zip file downloaded locally to {file_path}.")
        else:
            return zipfile.ZipFile(io.BytesIO(req.content))

    def get_route(self, method, **kwargs):
        """
        Helper function to automatically match and supply required arguments
        """
        route_method = getattr(routes, method)
        route_method_args, _, _, route_defaults = inspect.getargspec(route_method)
        if route_defaults:
            route_method_args = route_method_args[: -len(route_defaults)]
        # Update loaded argument dict with new arguments which are not None
        new_args = {k: v for k, v in list(kwargs.items()) if v is not None}
        arg_dict = dict(self.env_args, **new_args)
        input_args = []
        for arg in route_method_args:
            if arg_dict[arg]:
                input_args.append(arg_dict[arg])
            else:
                raise Exception(
                    f"For route: {method}, argument {arg} needs to be provided."
                )
        return route_method(  # pylint: disable=no-value-for-parameter
            *tuple(input_args)
        )

    def get(self, route, **kwargs):
        return self._call("get", route, **kwargs)

    def put(self, route, **kwargs):
        return self._call("put", route, **kwargs)

    def post(self, route, **kwargs):
        return self._call("post", route, **kwargs)

    def delete(self, route, **kwargs):
        return self._call("delete", route, **kwargs)

    def _req_call(self, method, route, **kwargs):
        return getattr(self.session, method)(route, **kwargs)

    def _call(self, method, route, status_response={}, **kwargs):
        """Base function for handling all requests"""
        if self.verbose:
            print(f"{method.upper()}: {route}")

        return self._handle_response(
            self._req_call(method, route, **kwargs), **status_response
        )

    def _handle_response(self, response, **kwargs):
        unauthorized_resp = (
            "You are not authorized to execute this command. "
            "For more information on access "
            "permissions see the package documentation."
        )
        internal_error_resp = (
            "An internal server error has occurred. "
            "Please contact support for assistance."
        )
        default_status_response = {
            "200": lambda resp: resp.json(),
            "201": lambda resp: resp.json(),
            "401": lambda resp: PermissionError(
                "[%d] %s" % (resp.status_code, unauthorized_resp)
            ),
            "403": lambda resp: PermissionError(
                "[%d] %s" % (resp.status_code, unauthorized_resp)
            ),
            "500": lambda resp: Exception(
                "[%d] %s" % (resp.status_code, internal_error_resp)
            ),
            "default": lambda resp: Exception(
                "[%d] %s" % (resp.status_code, resp.text)
            ),
        }
        status_response = dict(default_status_response, **kwargs)

        return_val = status_response.get(
            str(response.status_code), status_response["default"]
        )

        if isinstance(return_val(response), Exception):
            raise return_val(response)
        else:
            return return_val(response)

    # NOTE(meawoppl) This is only called externally
    def _post_analytics(self, client_id=None, event_action=None, event_category="cli"):
        route = "https://www.google-analytics.com/collect"
        if not client_id:
            client_id = self.user_id
        packet = (
            f"v=1&tid=UA-28937242-7&cid={client_id}&t=event&"
            f"ea={event_action}&ec={event_category}"
        )
        requests.post(route, packet)


class AnalysisException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


def _parse_protocol(protocol):
    if isinstance(protocol, dict):
        return protocol
    try:
        from autoprotocol import Protocol
    except ImportError:
        raise RuntimeError(
            "Please install `autoprotocol-python` in order "
            "to work with Protocol objects"
        )
    if isinstance(protocol, Protocol):
        return protocol.as_dict()



import json
import ast
import re
from collections import OrderedDict


PLURAL_UNITS = [
    "microliter",
    "nanoliter",
    "milliliter",
    "second",
    "minute",
    "hour",
    "g",
    "nanometer",
]

TEMP_DICT = {
    "cold_20": "-20 degrees celsius",
    "cold_80": "-80 degrees celsius",
    "warm_37": "37 degrees celsius",
    "cold_4": "4 degrees celsius",
    "warm_30": "30 degrees celsius",
    "ambient": "room temperature",
}


class AutoprotocolParser(object):
    def __init__(self, protocol_obj, api=None, parsed_output=None):
        self.api = api
        self.resource = dict()
        self.parse(protocol_obj)

    def parse(self, obj):
        self.object_list = []
        self.instructions = obj["instructions"]

        parsed_output = []
        for i in self.instructions:
            try:
                output = getattr(self, i["op"])(i)
                parsed_output.extend(output) if isinstance(
                    output, list
                ) else parsed_output.append(output)
            except AttributeError:
                parsed_output.append("[Unknown instruction]")

        self.parsed_output = parsed_output
        for i, p in enumerate(parsed_output):
            print("%d. %s" % (i + 1, p))

    def job_tree(self):
        """
        A Job Tree visualizes the instructions of a protocol in a hierarchical
        structure based on container dependency to help human readers with manual
        execution. Its construction utilizes the algorithm below, as well as the
        Node object class (to store relational information) at the bottom of this
        script.

        Example Usage:
            .. code-block:: python

                p = Protocol()

                bacterial_sample = p.ref("bacteria", None, "micro-1.5", discard=True)
                test_plate = p.ref("test_plate", None, "96-flat", storage="cold_4")

                p.dispense_full_plate(test_plate, "lb-broth-noAB", "50:microliter")
                w = 0
                amt = 1
                while amt < 20:
                    p.transfer(bacterial_sample.well(
                        0), test_plate.well(w), "%d:microliter" % amt)
                    amt += 2
                    w += 1

                pjsonString = json.dumps(p.as_dict(), indent=2)
                pjson = json.loads(pjsonString)
                parser_instance = english.AutoprotocolParser(pjson)
                parser_instance.job_tree()

                Output:
                1
                +---2
                3
                +---4
                5
                +---6
                7
                +---8
                9
                +---10
                11
                +---12


        Variables
        ---------
        steps: list
            deep list of objects per instruction/step;
            is primary information job tree is built from
        nodes: list
            list of node objects
        proto_forest: list
            list of lists grouped by connected nodes
        forest: list
            list of nested dictionaries, depicting parent-children relations
        forest_list: list
            list of nested lists, depticting parent-children relations
        """

        # 1. Enforce depth of 1 for steps
        def depth_one(steps):
            depth_one = []
            for step in steps:
                if type(step) is list:
                    if type(step[0]) is list:
                        depth_one.append(step[0])
                    else:
                        depth_one.append(step)
                else:
                    depth_one.append([step])
            return depth_one

        # 2. Convert steps to list of node objects (0,1,2,3...)
        def assign_nodes(steps):
            nodes = [i for i in range(len(steps))]
            objects = list(set([elem for sublist in steps for elem in sublist]))

            # checks for multiple src and dst objects -- added when looking for
            # mutiples
            split_objects = []
            for obj in objects:
                if len(obj) > 1:
                    new_objs = obj.split(", ")
                    split_objects.extend(new_objs)
                else:
                    split_objects.append(obj)
            objects = split_objects
            del split_objects

            # populate with leafless trees (Node objects, no edges)
            for node in nodes:
                nodes[node] = Node(str(node))

            # search for leafy trees
            for obj in objects:

                # accounts for multiple drc/dst objects
                leaves = []
                for i, sublist in enumerate(steps):
                    for string in sublist:
                        if string.count(",") > 0:
                            if obj in string:
                                leaves.append(i)
                        else:
                            if obj in sublist:
                                leaves.append(i)
                leaves = sorted(list(set(leaves)))

                if len(leaves) > 1:
                    viable_edges = []

                    # compute cross-product
                    for leaf1 in leaves:
                        for leaf2 in leaves:
                            if (
                                str(leaf1) != str(leaf2)
                                and sorted((leaf1, leaf2)) not in viable_edges
                            ):
                                viable_edges.append(sorted((leaf1, leaf2)))

                    # form edge networks
                    for edge in viable_edges:
                        n1, n2 = nodes[edge[0]], nodes[edge[1]]
                        n1.add_edge(n2)
                        n2.add_edge(n1)
                        nodes[int(n1.name)], nodes[int(n2.name)] = n1, n2
            return nodes

        # 3. Determine number of trees and regroup by connected nodes
        def connected_nodes(nodes):
            proto_trees = []
            nodes = set(nodes)

            while nodes:
                n = nodes.pop()
                group = {n}
                queue = [n]
                while queue:
                    n = queue.pop(0)
                    neighbors = n.edges
                    neighbors.difference_update(group)
                    nodes.difference_update(neighbors)
                    group.update(neighbors)
                    queue.extend(neighbors)
                proto_trees.append(group)
            return proto_trees

        # 4. Convert nodes to nested dictionary of parent-children relations
        # i.e. adding depth -- also deals with tree-node sorting and path
        # optimization
        def build_tree_dict(trees, steps):
            # node sorting in trees
            sorted_trees = []
            for tree in trees:
                sorted_trees.append(sorted(tree, key=lambda x: int(x.name)))

            # retrieve values of the nodes (the protocol's containers)
            # for each tree ... may want to use dictionary eventually
            all_values = []
            for tree in sorted_trees:
                values = [steps[int(node.name)] for node in tree]
                all_values.append(values)

            # create relational tuples:
            all_digs = []
            singles = []
            dst_potentials = []
            for tree_idx in range(len(sorted_trees)):
                edge_flag = False
                tree_digs = []
                for node_idx in range(len(sorted_trees[tree_idx])):

                    # digs: directed graph vectors
                    digs = []
                    dst_nodes = []
                    node_values = all_values[tree_idx][node_idx]
                    src_node = str(sorted_trees[tree_idx][node_idx].name)

                    # ACTION ON MULTIPLE OBJECTS (E.G. TRANSFER FROM SRC -> DST
                    # WELLS)
                    # Outcome space: {1-1, 1-many, many-1, many-many}
                    if len(node_values) == 2:
                        # single destination (x-1)
                        if node_values[1].count(",") == 0:
                            dst_nodes = [
                                i
                                for i, sublist in enumerate(steps)
                                if node_values[1] == sublist[0]
                            ]
                        # multiple destinations (x-many)
                        elif node_values[1].count(",") > 0:
                            dst_nodes = []
                            for dst in node_values[1].replace(", ", ""):
                                for i, sublist in enumerate(steps):
                                    if i not in dst_nodes and dst == sublist[0]:
                                        dst_nodes.append(i)

                    # ACTION ON A SINGLE OBJECT
                    elif len(node_values) == 1:
                        dst_nodes = [
                            i
                            for i, sublist in enumerate(steps)
                            if node_values[0] == sublist[0]
                        ]

                    # Constructing tuples in (child, parent) format
                    for dst_node in dst_nodes:
                        dig = (int(dst_node), int(src_node))
                        digs.append(dig)

                    # else: an edge-case for dictionaries constructed with no edges
                    # initiates tree separation via flag
                    if digs != []:
                        edge_flag = False
                        tree_digs.append(digs)
                    else:
                        edge_flag = True
                        digs = [(int(src_node), int(src_node))]
                        tree_digs.append(digs)

                # digraph cycle detection: avoids cycles by overlooking set
                # repeats
                true_tree_digs = []
                for digs in tree_digs:
                    for dig in digs:
                        if tuple(sorted(dig, reverse=True)) not in true_tree_digs:
                            true_tree_digs.append(tuple(sorted(dig, reverse=True)))

                # edge-case for dictionaries constructed with no edges
                if true_tree_digs != [] and edge_flag == False:
                    all_digs.append(true_tree_digs)
                elif edge_flag == True:
                    all_digs.extend(tree_digs)

            # Enforces forest ordering
            all_digs = sorted(all_digs, key=lambda x: x[0])

            # job tree traversal to find all paths:
            forest = []
            for digs_set in all_digs:

                # pass 1: initialize nodes dictionary
                nodes = OrderedDict()
                for tup in digs_set:
                    id, parent_id = tup
                    # ensure all nodes accounted for
                    nodes[id] = OrderedDict({"id": id})
                    nodes[parent_id] = OrderedDict({"id": parent_id})

                # pass 2: create trees and parent-child relations
                for tup in digs_set:
                    id, parent_id = tup
                    node = nodes[id]
                    # links node to its parent
                    if id != parent_id:
                        # add new_node as child to parent
                        parent = nodes[parent_id]
                        if not "children" in parent:
                            # ensure parent has a 'children' field
                            parent["children"] = []
                        children = parent["children"]
                        children.append(node)

                desired_tree_idx = sorted(list(nodes.keys()))[0]
                forest.append(nodes[desired_tree_idx])
            return forest

        # 5. Convert dictionary-stored nodes to unflattened, nested list of
        # parent-children relations
        def dict_to_list(forest):
            forest_list = []
            for tree in forest:
                tString = str(json.dumps(tree))
                tString = (
                    tString.replace('"id": ', "")
                    .replace('"children": ', "")
                    .replace("[{", "[")
                    .replace("}]", "]")
                    .replace("{", "[")
                    .replace("}", "]")
                )

                # find largest repeated branch (if applicable)
                # maybe think about using prefix trees or SIMD extensions for better
                # efficiency
                x, y, length, match = 0, 0, 0, ""
                for y in range(len(tString)):
                    for x in range(len(tString)):
                        substring = tString[y:x]
                        if (
                            len(list(re.finditer(re.escape(substring), tString))) > 1
                            and len(substring) > length
                        ):
                            match = substring
                            length = len(substring)

                # checking for legitimate branch repeat
                if "[" in match and "]" in match:
                    hits = []
                    index = 0
                    if len(tString) > 3:
                        while index < len(tString):
                            index = tString.find(str(match), index)
                            if index == -1:
                                break
                            hits.append(index)
                            index += len(match)

                    # find all locations of repeated branch and remove
                    if len(hits) > 1:
                        for start_loc in hits[1:]:
                            tString = tString[:start_loc] + tString[start_loc:].replace(
                                match, "]", 1
                            )

                # increment all numbers in string to match the protocol
                newString = ""
                numString = ""
                for el in tString:
                    if el.isdigit():  # build number
                        numString += el
                    else:
                        if (
                            numString != ""
                        ):  # convert it to int and reinstantaite numString
                            numString = str(int(numString) + 1)
                        newString += numString
                        newString += el
                        numString = ""
                tString = newString
                del newString

                forest_list.append(ast.literal_eval(tString))
            return forest_list

        # 6. Print job tree(s)
        def print_tree(lst, level=0):
            print("    " * (level - 1) + "+---" * (level > 0) + str(lst[0]))
            for l in lst[1:]:
                if type(l) is list:
                    print_tree(l, level + 1)
                else:
                    print("    " * level + "+---" + l)

        # 1
        steps = depth_one(self.object_list)
        # 2
        nodes = assign_nodes(steps)
        # 3
        proto_forest = connected_nodes(nodes)
        # 4
        forest = build_tree_dict(proto_forest, steps)
        # 5
        self.forest_list = dict_to_list(forest)
        # 6
        print("\n" + "A suggested Job Tree based on container dependency: \n")
        for tree_list in self.forest_list:
            print_tree(tree_list)

    def absorbance(self, opts):
        self.object_list.append([opts["object"]])
        return (
            f"Measure absorbance at {self.unit(opts['wavelength'])} for "
            f"{self.well_list(opts['wells'])} of plate {opts['object']}"
        )

    def acoustic_transfer(self, opts):
        transfers = []
        for t in opts["groups"][0]["transfer"]:
            transfers.append(
                f"Acoustic transfer {self.unit(t['volume'])} "
                f"from {t['from']} to {t['to']}"
            )
            self.object_list.append([t["from"], t["to"]])
        return transfers

    def autopick(self, opts):
        picks = []
        for i, g in enumerate(opts["groups"]):
            picks.extend(
                [
                    "Pick %s colonies from %s %s: %s to %s, %s"
                    % (
                        len(g["to"]),
                        len(g["from"]),
                        ("well" if len(g["from"]) is 1 else "wells"),
                        self.well_list(g["from"]),
                        self.well_list(g["to"]),
                        (
                            f"data saved at '{opts['dataref']}'"
                            if i is 0
                            else "analyzed with previous"
                        ),
                    )
                ]
            )
            self.object_list.append([g["from"], g["to"]])
        return picks

    def cover(self, opts):
        self.object_list.append([opts["object"]])
        return f"Cover {opts['object']} with a {opts['lid']} lid"

    def dispense(self, opts):
        self.object_list.append([opts["object"]])
        unique_vol = []
        for col in opts["columns"]:
            vol = self.unit(col["volume"])
            if vol not in unique_vol:
                unique_vol.append(vol)
        if "reagent" in opts:
            reagent = opts["reagent"]
        elif "resource_id" in opts:
            resource_id = opts["resource_id"]
            if resource_id in self.resource:
                reagent = self.resource[resource_id]
            elif self.api:
                resource = self.api.resources(resource_id)
                if resource["results"]:
                    reagent = resource["results"][0]["name"].lower()
                    self.resource[resource_id] = reagent
                else:
                    reagent = f"resource with resource ID {resource_id}"
            else:
                reagent = f"resource with resource ID {resource_id}"
        else:
            reagent = "unknown"

        if len(opts["columns"]) == 12 and len(unique_vol) == 1:
            return (
                f"Dispense {unique_vol[0]} of {reagent} to the full plate "
                f"of {opts['object']}"
            )
        else:
            return "Dispense corresponding amounts of %s to %d column(s) of %s" % (
                reagent,
                len(opts["columns"]),
                opts["object"],
            )

    def flash_freeze(self, opts):
        self.object_list.append([opts["object"]])
        return f"Flash freeze {opts['object']} for {self.unit(opts['duration'])}"

    def fluorescence(self, opts):
        self.object_list.append([opts["object"]])
        return (
            f"Read fluorescence of {self.well_list(opts['wells'])} of "
            f"plate {opts['object']} at excitation wavelength "
            f"{self.unit(opts['excitation'])} and emission wavelength "
            f"{self.unit(opts['emission'])}"
        )

    def gel_separate(self, opts):
        self.object_list.append([opts["matrix"]])
        return (
            f"Perform gel electrophoresis using a "
            f"{opts['matrix'].split(',')[1][:-1]} agarose gel for "
            f"{self.unit(opts['duration'])}"
        )

    def gel_purify(self, opts):
        self.object_list.append([opts["matrix"]])
        unique_bl = []
        for ext in opts["extract"]:
            bl = ext["band_size_range"]
            if bl not in unique_bl:
                unique_bl.append(bl)
        for i in range(len(unique_bl)):
            unique_bl[i] = (
                str(unique_bl[i]["min_bp"]) + "-" + str(unique_bl[i]["max_bp"])
            )

        if len(unique_bl) <= 3:
            return (
                f"Perform gel purification on the "
                f"{opts['matrix'].split(',')[1][:-1]} agarose gel with "
                f"band range(s) {', '.join(unique_bl)}"
            )
        else:
            return (
                f"Perform gel purification on the "
                f"{opts['matrix'].split(',')[1][:-1]} agarose gel with "
                f"{len(unique_bl)} band ranges"
            )

    def incubate(self, opts):
        self.object_list.append([opts["object"]])
        shaking = " (shaking)" if opts["shaking"] else ""
        return (
            f"Incubate {opts['object']} at "
            f"{TEMP_DICT[opts['where']]} for "
            f"{self.unit(opts['duration'])}{shaking}"
        )

    def image_plate(self, opts):
        self.object_list.append([opts["object"]])
        return f"Take an image of {opts['object']}"

    def luminescence(self, opts):
        self.object_list.append([opts["object"]])
        return (
            f"Read luminescence of {self.well_list(opts['wells'])} of "
            f"plate {opts['object']}"
        )

    def oligosynthesize(self, opts):
        self.object_list.append([o["destination"] for o in opts["oligos"]])
        return [
            f"Oligosynthesize sequence '{o['sequence']}' into " f"'{o['destination']}'"
            for o in opts["oligos"]
        ]

    def provision(self, opts):
        self.object_list.append([self.platename(t["well"]) for t in opts["to"]])
        resource_id = opts["resource_id"]
        if resource_id in self.resource:
            reagent = self.resource[resource_id]
        elif self.api:
            resource = self.api.resources(resource_id)
            if resource["results"]:
                reagent = resource["results"][0]["name"].lower()
                self.resource[resource_id] = reagent
            else:
                reagent = f"resource with resource ID {resource_id}"
        else:
            reagent = f"resource with resource ID {resource_id}"
        provisions = []
        for t in opts["to"]:
            provisions.append(
                f"Provision {self.unit(t['volume'])} of "
                f"{reagent} to well {self.well(t['well'])} of "
                f"container {self.platename(t['well'])}"
            )
        return provisions

    def sanger_sequence(self, opts):
        self.object_list.append([opts["object"]])
        seq = (
            f"Sanger sequence {self.well_list(opts['wells'])} of plate "
            f"{opts['object']}"
        )
        if opts["type"] == "standard":
            return seq
        elif opts["type"] == "rca":
            return seq + " with %s" % self.platename(opts["primer"])

    def illumina_sequence(self, opts):
        unique_wells = self.get_unique_wells(opts["lanes"])
        unique_plates = self.get_unique_plates(unique_wells)
        self.object_list.append(unique_plates)

        if len(unique_plates) == 1 and len(unique_wells) <= 3:
            seq = "Illumina sequence wells %s" % (", ".join(unique_wells))
        elif len(unique_plates) > 1 and len(unique_plates) <= 3:
            seq = "Illumina sequence the corresponding wells of plates %s" % ", ".join(
                unique_plates[0]
            )
        else:
            seq = "Illumina sequence the corresponding wells of %s plates" % len(
                unique_wells
            )

        return seq + f" with library size {opts['library_size']}"

    def flow_analyze(self, opts):
        wells = []
        for sample in opts["samples"]:
            if sample["well"] not in wells:
                wells.append(sample["well"])
        self.object_list.append([self.platename(w) for w in wells])

        return (
            "Perform flow cytometry on %s with the respective FSC and SSC channel parameters"
            % ", ".join(wells)
        )

    def seal(self, opts):
        self.object_list.append([opts["object"]])
        return f"Seal {opts['object']} ({opts['type']})"

    def spin(self, opts):
        self.object_list.append([opts["object"]])
        return (
            f"Spin {opts['object']} for {self.unit(opts['duration'])} at "
            f"{self.unit(opts['acceleration'])}"
        )

    def spread(self, opts):
        self.object_list.append([self.well(opts["from"]), self.well(opts["to"])])
        return [
            f"Spread {opts['volume']} of bacteria from well "
            f"{self.well(opts['from'])} of {self.platename(opts['from'])} "
            f"to well {self.well(opts['to'])} of agar plate "
            f"{self.platename(opts['to'])}"
        ]

    def stamp(self, opts):
        stamps = []
        for g in opts["groups"]:
            for pip in g:
                if pip == "transfer":
                    stamps.extend(
                        [
                            "Stamp %s from source origin %s "
                            "to destination origin %s %s (%s)"
                            % (
                                self.unit(p["volume"]),
                                p["from"],
                                p["to"],
                                (
                                    "with the same set of tips as previous"
                                    if (len(g[pip]) > 1 and i > 0)
                                    else ""
                                ),
                                (
                                    "%s rows x %s columns"
                                    % (g["shape"]["rows"], g["shape"]["columns"])
                                ),
                            )
                            for i, p in enumerate(g[pip])
                        ]
                    )
                    from_objs = str(
                        [self.platename(p["from"]) for i, p in enumerate(g[pip])]
                    )
                    to_objs = str(
                        [self.platename(p["to"]) for i, p in enumerate(g[pip])]
                    )
                    self.object_list.append([from_objs, to_objs])
        return stamps

    def thermocycle(self, opts):
        self.object_list.append([opts["object"]])
        return f"Thermocycle {opts['object']}"

    def pipette(self, opts):
        pipettes = []
        for g in opts["groups"]:
            for pip in g:
                if pip == "mix":
                    for m in g[pip]:
                        pipettes.append(
                            "Mix well %s of plate %s %d times "
                            "with a volume of %s"
                            % (
                                self.well(m["well"]),
                                self.platename(m["well"]),
                                m["repetitions"],
                                self.unit(m["volume"]),
                            )
                        )
                        self.object_list.append(self.platename(m["well"]))
                elif pip == "transfer":
                    pipettes.extend(
                        [
                            f"Transfer {self.unit(p['volume'])} from {p['from']} to {p['to']} {'with the same tip as previous' if len(g[pip]) > 1 and i > 0 else ''}"
                            for i, p in enumerate(g[pip])
                        ]
                    )

                    from_objs = str(
                        [self.platename(p["from"]) for i, p in enumerate(g[pip])]
                    )
                    to_objs = str(
                        [self.platename(p["to"]) for i, p in enumerate(g[pip])]
                    )
                    self.object_list.append([from_objs, to_objs])
                elif pip == "distribute":
                    pipettes.append(
                        f"Distribute from {g[pip]['from']} into "
                        f"{self.well_list([d['well'] for d in g[pip]['to']], 20)}"
                    )
                    self.object_list.append([g[pip]["from"], g[pip]["to"][0]["well"]])
                elif pip == "consolidate":
                    pipettes.append(
                        f"Consolidate "
                        f"{self.well_list([c['well'] for c in g[pip]['from']], 20)} "
                        f"into {g[pip]['to']}"
                    )
                    self.object_list.append([g[pip]["from"][0]["well"], g[pip]["to"]])
        return pipettes

    def magnetic_transfer(self, opts):
        specific_op = list(opts["groups"][0][0].keys())[0]
        specs_dict = opts["groups"][0][0][specific_op]
        self.object_list.append([specs_dict["object"]])
        seq = f"Magnetically {specific_op} {specs_dict['object']}"

        if specific_op == "dry":
            return seq + " for %s" % self.unit(specs_dict["duration"])
        elif specific_op == "incubate":
            return (
                seq + f" for {self.unit(specs_dict['duration'])} with a "
                f"tip position of {specs_dict['tip_position']}"
            )
        elif specific_op == "collect":
            return (
                seq + f" beads for {specs_dict['cycles']} cycles with a "
                f"pause duration of {self.unit(specs_dict['pause_duration'])}"
            )
        elif specific_op == "release" or "mix":
            return (
                seq + f" beads for {self.unit(specs_dict['duration'])} at "
                f"an amplitude of {specs_dict['amplitude']}"
            )

    def measure_volume(self, opts):
        unique_plates = self.get_unique_plates(opts["object"])
        self.object_list.append(unique_plates)

        if len(unique_plates) <= 3:
            return f"Measure volume of {len(opts['object'])} wells from {', '.join(unique_plates)}"
        else:
            return f"Measure volume of {len(opts['object'])} wells from the {len(unique_plates)} plates"

    def measure_mass(self, opts):
        unique_plates = self.get_unique_plates(opts["object"])
        self.object_list.append(unique_plates)
        return "Measure mass of %s" % ", ".join(opts["object"])

    def measure_concentration(self, opts):
        unique_plates = self.get_unique_plates(opts["object"])
        self.object_list.append(unique_plates)
        return (
            f"Measure concentration of {self.unit(opts['volume'])} "
            f"{opts['measurement']} source aliquots of {self.platename(opts['object'][0])}"
        )

    def uncover(self, opts):
        self.object_list.append([opts["object"]])
        return f"Uncover {opts['object']}"

    def unseal(self, opts):
        self.object_list.append([opts["object"]])
        return f"Unseal {opts['object']}"

    @staticmethod
    def platename(ref):
        return ref.split("/")[0]

    @staticmethod
    def well(ref):
        return ref.split("/")[1]

    @staticmethod
    def get_unique_wells(list_of_wells):
        unique_wells = []
        for well in list_of_wells:
            w = well["object"]
            if w not in unique_wells:
                unique_wells.append(w)
        return unique_wells

    @staticmethod
    def get_unique_plates(list_of_wells):
        unique_plates = []
        for well in list_of_wells:
            loc = well.find("/")
            if loc == -1:
                plate = well
            else:
                plate = well[:loc]

            if plate not in unique_plates:
                unique_plates.append(plate)
        return unique_plates

    @staticmethod
    def well_list(wells, max_len=10):
        well_list = "wells " + (", ").join(str(x) for x in wells)
        if len(wells) > max_len:
            well_list = str(len(wells)) + " wells"
        return well_list

    @staticmethod
    def unit(u):
        value = u.split(":")[0]
        unit = u.split(":")[1]
        return (
            f"{value} "
            f"{unit + 's' if float(value) > 1 and unit in PLURAL_UNITS else unit}"
        )


class Node(object):
    """
    A Node represents a Job Tree element that fulfils a broader child-parent
    relational structure. It contains relevant information on its relationships
    in the form of edges. The job_tree algorithm above then constructs the
    actual hierachy of the aforementioned relational structure.

    Example Usage:
        .. code-block:: python

        a = Node("a")
        b = Node("b")
        c = Node("c")
        d = Node("d")
        e = Node("e")
        f = Node("f")      structure:
        a.add_edge(b)          a
        a.add_edge(c)         / \
        b.add_edge(d)        b   c
        c.add_edge(e)       /   / \
        c.add_edge(f)      d   e   f

    Attributes
    ----------
    name: str
        Name the Node object
    edges: set
        Set of edges that each node owns
    """

    def __init__(self, name):
        self.__name = name
        self.__links = set()

    @property
    def name(self):
        return self.__name

    @property
    def edges(self):
        return set(self.__links)

    def add_edge(self, other):
        self.__links.add(other)
        other.__links.add(self)



#!/usr/bin/env python3

import click
import os
import requests

from transcriptic.config import Connection
from transcriptic import commands


class FeatureGroup(click.Group):
    """Custom group to handle hiding of commands based on the `feature` tag
    TODO: Deprecate once Click 7 lands and use `hidden` parameter in commands
    """

    def __init__(self, **attrs):
        click.Group.__init__(self, **attrs)

    def format_commands(self, ctx, formatter):
        """Custom formatter to control whether a command is displayed
        Note: This is only called when formatting the help message.
        """
        ctx.obj = ContextObject()
        try:
            ctx.obj.api = Connection.from_file("~/.transcriptic")
        except (FileNotFoundError, OSError):
            # This defaults to feature_groups = []
            ctx.obj.api = Connection()

        rows = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if cmd is None:
                continue
            try:
                if (
                    cmd.feature is not None
                    and cmd.feature in ctx.obj.api.feature_groups
                ):
                    help = cmd.short_help or ""
                    rows.append((subcommand, help))
                else:
                    continue
            except AttributeError:
                help = cmd.short_help or ""
                rows.append((subcommand, help))

        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)


class FeatureCommand(click.Command):
    """Extend off Command to add `feature` attribute
    TODO: Deprecate once Click 7 lands and use `hidden` parameter in commands
    """

    def __init__(self, feature=None, **attrs):
        click.Command.__init__(self, **attrs)
        self.feature = feature


class HiddenOption(click.Option):
    """Monkey patch of click Option to enable hidden options
    TODO: Deprecate once Click 7 lands and use `hidden` option instead
    """

    def __init__(self, *param_decls, **attrs):
        __hidden__ = attrs.pop("hidden", True)
        click.Option.__init__(self, *param_decls, **attrs)
        self.__hidden__ = __hidden__

    def get_help_record(self, ctx):
        """This hijacks the help record so that a hidden option does not show
        up in the help text
        """
        if self.__hidden__:
            return
        click.Option.get_help_record(self, ctx)


class ContextObject(object):
    """Object passed along Click context
    Note: `ctx` is passed along whenever the @click.pass_context decorator is
    present. This object is referenced using `ctx.obj`
    """

    def __init__(self):
        self._api = None

    @property
    def api(self):
        return self._api

    @api.setter
    def api(self, value):
        self._api = value


_CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=_CONTEXT_SETTINGS, cls=FeatureGroup)
@click.option("--api-root", default=None, hidden=True, cls=HiddenOption)
@click.option("--email", default=None, hidden=True, cls=HiddenOption)
@click.option("--token", default=None, hidden=True, cls=HiddenOption)
@click.option("--organization", "-o", default=None, hidden=True, cls=HiddenOption)
@click.option(
    "--config",
    envvar="TRANSCRIPTIC_CONFIG",
    default="~/.transcriptic",
    help="Specify a configuration file.",
)
@click.version_option(prog_name="Transcriptic Python Library (TxPy)")
@click.pass_context
def cli(ctx, api_root, email, token, organization, config):
    """A command line tool for working with Transcriptic.

    Note: This is the main entry point of the CLI. If specifying credentials,
    note that the order of preference is: --flag, environment then config file.

    Example: `transcriptic --organization "my_org" projects` >>
    `export USER_ORGANIZATION="my_org"` >> `"organization_id": "my_org" in
     ~/.transcriptic
    """
    # Initialize ContextObject to be used for storing api object
    ctx.obj = ContextObject()

    if ctx.invoked_subcommand in ["compile", "preview", "summarize", "init"]:
        # For local commands, initialize empty connection
        ctx.obj.api = Connection()
    elif ctx.invoked_subcommand == "login":
        # Load analytics option from existing dotfile if present, else prompt
        try:
            api = Connection.from_file(config)
            api.api_root = api_root or os.environ.get("BASE_URL", None) or api.api_root
            ctx.obj.api = api
        except (OSError, IOError):
            ctx.obj.api = Connection()
        # Echo a warning if other options are defined for login
        if organization or email or token:
            click.echo(
                "Only the `--api-root` option is applicable for the "
                "`login` command. All other options are ignored."
            )
    else:
        try:
            api = Connection.from_file(config)
            api.api_root = api_root or os.environ.get("BASE_URL", None) or api.api_root
            api.organization_id = (
                organization
                or os.environ.get("USER_ORGANIZATION", None)
                or api.organization_id
            )
            api.email = email or os.environ.get("USER_EMAIL", None) or api.email
            api.token = token or os.environ.get("USER_TOKEN", None) or api.token
            ctx.obj.api = api
        except (OSError, IOError):
            click.echo(
                "Welcome to TxPy! It seems like your `.transcriptic` "
                "config file is missing or out of date"
            )
            analytics = click.confirm(
                "Send TxPy CLI usage information to "
                "improve the CLI user "
                "experience?",
                default=True,
            )
            ctx.obj.api = Connection()  # Initialize empty connection
            ctx.invoke(login_cmd, api_root=api_root, analytics=analytics)
    if ctx.obj.api.analytics:
        try:
            ctx.obj.api._post_analytics(
                event_action=ctx.invoked_subcommand, event_category="cli"
            )
        except requests.exceptions.RequestException:
            pass


@cli.command("submit", cls=FeatureCommand, feature="can_submit_autoprotocol")
@click.argument("file", default="-")
@click.option(
    "--project",
    "-p",
    metavar="PROJECT_ID",
    required=True,
    help=(
        "Project id or name to submit the run to. "
        "Use `transcriptic projects` command to list existing projects."
    ),
)
@click.option("--title", "-t", help="Optional title of your run")
@click.option("--test", help="Submit this run in test mode", is_flag=True)
@click.option(
    "--pm",
    metavar="PAYMENT_METHOD_ID",
    required=False,
    help="Payment id to be used for run submission. "
    "Use `transcriptic payments` command to list existing "
    "payment methods.",
)
@click.pass_context
def submit_cmd(ctx, file, project, title=None, test=None, pm=None):
    """Submit your run to the project specified."""
    api = ctx.obj.api
    commands.submit(api, file, project, title=title, test=test, pm=pm)


@cli.command("build-release", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("package", required=False, metavar="PACKAGE")
@click.option("--name", "-n", help="Optional name for your zip file")
@click.pass_context
def release_cmd(ctx, name=None, package=None):
    """ Compress the contents of the current directory to upload as a release.  """
    api = ctx.obj.api
    commands.release(api, name=name, package=package)


@cli.command("upload-release", cls=FeatureCommand, feature="can_upload_packages")
@click.argument(
    "archive", required=True, type=click.Path(exists=True), metavar="ARCHIVE"
)
@click.argument("package", required=True, metavar="PACKAGE")
@click.pass_context
def upload_release_cmd(ctx, archive, package):
    """Upload a release archive to a package."""
    api = ctx.obj.api
    commands.upload_release(api, archive, package)


@cli.command("upload-dataset")
@click.argument("file_path", type=click.Path(exists=True), metavar="FILE")
@click.argument("title", metavar="TITLE")
@click.argument("run_id", metavar="RUN-ID")
@click.option(
    "--tool",
    "-t",
    required=True,
    help="Name of analysis tool used for generating the dataset",
)
@click.option(
    "--version",
    "-v",
    required=True,
    help="Version of analysis tool used for generating the dataset",
)
@click.pass_context
def upload_dataset_cmd(ctx, file_path, title, run_id, tool, version):
    """Uploads specified file as an analysis dataset to the specified run."""
    api = ctx.obj.api
    commands.upload_dataset(api, file_path, title, run_id, tool, version)


@cli.command("protocols")
@click.pass_context
@click.option(
    "--local",
    is_flag=True,
    required=False,
    default=False,
    help="Shows available local protocols instead of remote protocols",
)
@click.option("--json", "json_flag", help="print JSON response", is_flag=True)
def protocols_cmd(ctx, local, json_flag):
    """List protocols within your manifest or organization."""
    api = ctx.obj.api
    commands.protocols(api, local, json_flag)


@cli.command("packages")
@click.pass_context
@click.option("-i")
def packages_cmd(ctx, i):
    """List packages in your organization."""
    api = ctx.obj.api
    commands.packages(api, i)


@cli.command("create-package", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("name")
@click.argument("description")
@click.pass_context
def create_package_cmd(ctx, description, name):
    """Create a new empty protocol package"""
    api = ctx.obj.api
    commands.create_package(api, description, name)


@cli.command("delete-package", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("name")
@click.option(
    "--force",
    "-f",
    help="force delete a package without being \
              prompted if you're sure",
    is_flag=True,
)
@click.pass_context
def delete_package_cmd(ctx, name, force):
    """Delete an existing protocol package"""
    api = ctx.obj.api
    commands.delete_package(api, name, force)


@cli.command("generate-protocol")
@click.pass_context
@click.argument("name")
def generate_protocol_cmd(ctx, name):
    """Generate a python protocol scaffold"""
    commands.generate_protocol(name)


@cli.command("projects")
@click.pass_context
@click.option("-i")
@click.option("--json", "json_flag", help="print JSON response", is_flag=True)
def projects_cmd(ctx, i, json_flag):
    """List the projects in your organization"""
    api = ctx.obj.api
    commands.projects(api, i, json_flag)


@cli.command("runs")
@click.pass_context
@click.argument("project_name")
@click.option("--json", "json_flag", help="print JSON response", is_flag=True)
def runs_cmd(ctx, project_name, json_flag):
    """List the runs that exist in a project"""
    api = ctx.obj.api
    commands.runs(api, project_name, json_flag)


@cli.command("create-project")
@click.argument("name", metavar="PROJECT_NAME")
@click.option("--dev", "-d", "-pilot", help="Create a pilot project", is_flag=True)
@click.pass_context
def create_project_cmd(ctx, name, dev):
    """Create a new empty project."""
    api = ctx.obj.api
    commands.create_project(api, name, dev)


@cli.command("delete-project")
@click.argument("name", metavar="PROJECT_NAME")
@click.option(
    "--force",
    "-f",
    help="force delete a project without being \
              prompted if you're sure",
    is_flag=True,
)
@click.pass_context
def delete_project_cmd(ctx, name, force):
    """Delete an existing project."""
    api = ctx.obj.api
    commands.delete_project(api, name, force)


@cli.command("resources")
@click.argument("query", default="*")
@click.pass_context
def resources_cmd(ctx, query):
    """Search catalog of provisionable resources"""
    api = ctx.obj.api
    commands.resources(api, query)


@cli.command("inventory")
@click.argument("query", default="*")
@click.option(
    "--include_aliquots", help="include containers with matching aliquots", is_flag=True
)
@click.option("--show_status", help="show container status", is_flag=True)
@click.option(
    "--retrieve_all", help="retrieve all samples, this may take a while", is_flag=True
)
@click.pass_context
def inventory_cmd(ctx, include_aliquots, show_status, retrieve_all, query):
    """Search organization for inventory"""
    api = ctx.obj.api
    commands.inventory(api, include_aliquots, show_status, retrieve_all, query)


@cli.command("payments")
@click.pass_context
def payments_cmd(ctx):
    """Lists available payment methods"""
    api = ctx.obj.api
    commands.payments(api)


@cli.command("init", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("path", default=".")
def init_cmd(path):
    """Initialize a directory with a manifest.json file."""
    commands.init(path)


@cli.command("analyze")
@click.argument("file", default="-")
@click.option("--test", help="Analyze this run in test mode", is_flag=True)
@click.pass_context
def analyze_cmd(ctx, file, test):
    """Analyze a block of Autoprotocol JSON."""
    api = ctx.obj.api
    commands.analyze(api, file, test)


@cli.command("preview", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("protocol_name", metavar="PROTOCOL_NAME")
@click.option("--view", is_flag=True)
@click.option("--dye_test", is_flag=True)
@click.pass_context
def preview_cmd(ctx, protocol_name, view, dye_test):
    """Preview the Autoprotocol output of protocol in the current package."""
    api = ctx.obj.api
    commands.preview(api, protocol_name, view, dye_test)


@cli.command("summarize")
@click.argument("file", default="-")
@click.pass_context
@click.option(
    "--html", "-x", is_flag=True, help="Generates an html view of the autoprotocol"
)
@click.option(
    "--tree",
    "-t",
    is_flag=True,
    help="Prints a job tree with instructions as leaf nodes",
)
@click.option(
    "--lookup",
    "-l",
    is_flag=True,
    help="Queries Transcriptic to convert resourceID to string",
)
# time allowance is on order of seconds
@click.option("--runtime", type=click.INT, default=5)
def summarize_cmd(ctx, file, html, tree, lookup, runtime):
    """Summarize Autoprotocol as a list of plain English steps, as well as a
    visualized Job Tree contingent upon desired runtime allowance (in seconds).
    A Job Tree refers to a structure of protocol based on container dependency,
    where each node, and its corresponding number, represents an instruction of
    the protocol. More specifically, the tree structure contains process branches,
    in which the x-axis refers to the dependency depth in a given branch, while
    the y-axis refers to the traversal of branches themselves.

    Example usage is as follows:

    python my_script.py | transcriptic summarize --tree

    python my_script.py | transcriptic summarize --tree --runtime 20

    python my_script.py | transcriptic summarize --html
    """
    if lookup or html:
        try:
            config = "~/.transcriptic"
            ctx.obj = ContextObject()
            ctx.obj.api = Connection.from_file(config)
        except:
            click.echo(
                "Connection with Transcriptic failed. Summarizing without lookup.",
                err=True,
            )

    api = ctx.obj.api
    commands.summarize(api, file, html, tree, lookup, runtime)


@cli.command("compile", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("protocol_name", metavar="PROTOCOL_NAME")
@click.argument("args", nargs=-1)
def compile_cmd(protocol_name, args):
    """Compile a protocol by passing it a config file (without submitting or
    analyzing)."""
    commands.compile(protocol_name, args)


@cli.command("launch")
@click.argument("protocol")
@click.argument(
    "params", metavar="PARAMETERS_FILE", type=click.File("r"), required=False
)
@click.option(
    "--project",
    "-p",
    metavar="PROJECT_ID",
    required=False,
    help="Project id or name context for configuring the protocol. Use "
    "`transcriptic projects` command to list existing projects.",
)
@click.option(
    "--title",
    "-t",
    metavar="RUN_TITLE",
    required=False,
    help="If specified, will apply custom title to run created, default run title"
    "will be the DISPLAY-NAME_MM_DD_YYYY of the protocol selected.",
)
@click.option(
    "--save_input",
    metavar="FILE",
    required=False,
    help="Save the protocol or parameters input JSON in a file. This is "
    "useful for debugging a protocol.",
)
@click.option(
    "--local",
    is_flag=True,
    required=False,
    help="If specified, the protocol will launch a local protocol and submit a run.",
)
@click.option(
    "--accept_quote",
    is_flag=True,
    required=False,
    help="If specified, the quote will automatically be accepted, and a run "
    "will be directly submitted.",
)
@click.option(
    "--pm",
    metavar="PAYMENT_METHOD_ID",
    required=False,
    help="Payment id to be used for run submission. "
    "Use `transcriptic payments` command to list existing "
    "payment methods.",
)
@click.option("--test", help="Submit this run in test mode", is_flag=True)
@click.option(
    "--pkg",
    metavar="PACKAGE_ID",
    required=False,
    help="Package ID for discriminating between protocols with identical names",
)
@click.pass_context
def launch_cmd(
    ctx,
    protocol,
    project,
    title,
    save_input,
    local,
    accept_quote,
    params,
    pm=None,
    test=None,
    pkg=None,
):
    """Configure and launch a protocol either using the local manifest file or remotely.
    If no parameters are specified, uses the webapp to select the inputs."""
    api = ctx.obj.api
    commands.launch(
        api,
        protocol,
        project,
        title,
        save_input,
        local,
        accept_quote,
        params,
        pm=None,
        test=None,
        pkg=None,
    )


@cli.command("select-org")
@click.argument("organization", metavar="ORGANIZATION_NAME", type=str, required=False)
@click.pass_context
def select_org_cmd(ctx, organization=None):
    """Allows you to switch organizations. If the organization argument
    is provided, this will directly select the specified organization.
    """
    api = ctx.obj.api
    config = ctx.parent.params["config"]
    commands.select_org(api, config, organization)


@cli.command("login")
@click.option(
    "--rsa-key",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Path to RSA key used for signing requests",
)
@click.pass_context
def login_cmd(ctx, api_root=None, analytics=True, rsa_key=None):
    """Authenticate to your Transcriptic account."""
    api = ctx.obj.api
    config = ctx.parent.params["config"]
    commands.login(api, config, api_root, analytics, rsa_key)


@cli.command("format", cls=FeatureCommand, feature="can_upload_packages")
@click.argument("manifest", default="manifest.json")
def format_cmd(manifest):
    """Check Autoprotocol format of manifest.json."""
    commands.format(manifest)



#!/usr/bin/env python3
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

# Load version
exec(open("transcriptic/version.py").read())


# Test Runner (reference: https://docs.pytest.org/en/latest/goodpractices.html)
class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = "--cov=transcriptic --cov-report=term"

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


# Test and Documentation dependencies
test_deps = [
    "coverage>=4.5, <5",
    "jsonschema>=2.6, <3",
    "mock>=3, <4",
    "pre-commit>=2.4, <3",
    "pylint==2.5.2",  # should be consistent with .pre-commit-config.yaml
    "pytest>=5.4, <6",
    "pytest-cov>=2, <3",
    "tox>=3.15, <4",
]

doc_deps = [
    "Sphinx>=2.4, <3",
    "mock>=3, <4",
    "releases>=1.6.3, <2",
    "sphinx_rtd_theme>=0.4.3, <1",
]

# Extra module dependencies
jupyter_deps = ["pandas>=0.23,<1"]

analysis_deps = [
    "autoprotocol>=7.1,<8",
    "matplotlib>=3,<4",
    "numpy>=1.14,<2",
    "pandas>=0.23,<1",
    "pillow>=3,<4",
    "plotly>=1.13,<2",
    "scipy>=0.14,<1",
]


setup(
    name="transcriptic",
    description="Transcriptic CLI & Python Client Library",
    url="https://github.com/transcriptic/transcriptic",
    version=__version__,  # pylint: disable=undefined-variable
    packages=["transcriptic", "transcriptic.jupyter", "transcriptic.analysis"],
    include_package_data=True,
    tests_require=test_deps,
    python_requires=">=3.6",
    install_requires=[
        "Click>=7.0,<8",
        "httpsig==1.3.0",
        "requests>=2.0,<3",
        "pycryptodome==3.9.6",
        "python-magic>=0.4,<1",
        "Jinja2>=2.0,<3",
    ],
    extras_require={
        "jupyter": jupyter_deps,
        "analysis": analysis_deps,
        "docs": doc_deps,
        "test": test_deps,
    },
    cmdclass={"pytest": PyTest},
    entry_points="""
        [console_scripts]
        transcriptic=transcriptic.cli:cli
    """,
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from io import BytesIO

from transcriptic import api

try:
    from PIL import Image
except ImportError:
    raise ImportError(
        "Please run `pip install transcriptic[analysis] if you "
        "would like to use the Transcriptic analysis module."
    )


class ImagePlate(object):
    """
    An ImagePlate object generalizes the parsing of datasets derived from the
    plate camera for easy visualization.

    Parameters
    ----------
    dataset: dataset
        Single dataset selected from datasets object

    Attributes
    ----------
    raw: BytesIO
        Raw buffer of image bytes
    image: PIL.Image
        Image object as rendered by PIL
    """

    def __init__(self, dataset):
        if dataset.attributes["instruction"]["operation"]["op"] != "image_plate":
            raise RuntimeError("No image_plate operation found for given dataset.")
        self.id = dataset.id
        self.raw = BytesIO()
        req = api.raw_image_data(data_id=self.id)
        # Buffer download of data
        chunk_sz = 512
        for chunk in req.iter_content(chunk_sz):
            self.raw.write(chunk)
        # Reset seek to 0
        self.raw.seek(0)
        self.image = Image.open(self.raw)

    def display(self):
        """
        Displays the original full-sized image. Helpful when used in an IPython
        kernel

        Returns
        -------
        HTML
            Returns a HTML iframe of the full-size image which is rendered nicely in IPython (if IPython is present)
        """
        try:
            # pylint: disable=import-error
            from IPython.display import HTML

            return HTML(
                """<iframe src="%s")" frameborder="0" \
                allowtransparency="true" style="height:500px;" seamless> \
                </iframe>"""
                % api.get_route("view_raw_image", data_id=self.id)
            )

        except:
            # If IPython module is not present or unable to show, display using
            # default PIL image show
            self.image.show()
